"""
 @file
 @brief This file sets the default logging settings
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os, sys
import copy
import logging
import logging.handlers

from classes import info


class StreamToLogger(object):
    """Custom class to log all stdout and stderr streams (from libopenshot / and other libraries)"""
    def __init__(self, parent_stream, log_level=logging.INFO):
        self.parent = parent_stream or sys.__stderr__
        self.logger = logging.LoggerAdapter(
            logging.getLogger('OpenShot.stderr'), {'source': 'stream'})
        self.log_level = log_level
        self.logbuf = ''

    def write(self, text):
        self.logbuf += str(text) or ""
        self.parent.write(text)

    def flush(self):
        if self.logbuf.rstrip():
            self.logger.log(self.log_level, self.logbuf.rstrip())
        self.logbuf = ''

    def errors(self):
        pass


class StreamFilter(logging.Filter):
    """Filter out lines that originated on the output"""
    def filter(self, record):
        source = getattr(record, "source", "")
        return source != "stream"

# Clamp log messages to a reasonable size to avoid giant lines
MAX_LOG_MESSAGE_LENGTH = 2048


class TruncatingFormatter(logging.Formatter):
    """Formatter that trims overly long log messages."""
    def format(self, record):
        message = record.getMessage()
        if len(message) > MAX_LOG_MESSAGE_LENGTH:
            record = copy.copy(record)
            record.msg = message[:MAX_LOG_MESSAGE_LENGTH] + "... [truncated]"
            record.args = None
        return super().format(record)


# Set up log formatters
template = '%(levelname)s %(module)s: %(message)s'
console_formatter = TruncatingFormatter(template)
file_formatter = TruncatingFormatter('%(asctime)s ' + template, datefmt='%H:%M:%S')

# Configure root logger for minimal logging
logging.basicConfig(level=logging.ERROR)
root_log = logging.getLogger()

# Set up our top-level logging context
log = root_log.getChild('OpenShot')
log.setLevel(info.LOG_LEVEL_FILE)
# Don't pass messages on to root logger
log.propagate = False

#
# Create rotating file handler
#
fh = None
if os.path.exists(info.USER_PATH):
    log_path = os.path.join(info.USER_PATH, 'openshot-qt.log')
    try:
        fh = logging.handlers.RotatingFileHandler(
            log_path, encoding="utf-8", maxBytes=25*1024*1024, backupCount=3
        )
    except OSError:
        # Fall back silently if the log file cannot be created (e.g. during tests
        # in read-only environments)
        fh = logging.NullHandler()

if fh:
    fh.setLevel(info.LOG_LEVEL_FILE)
    fh.setFormatter(file_formatter)
    # Only add the handler when it's a real logger (NullHandler is harmless)
    log.addHandler(fh)
else:
    fh = logging.NullHandler()

#
# Create typical stream handler which logs to stderr
#
sh = logging.StreamHandler(sys.stderr)
sh.setLevel(info.LOG_LEVEL_CONSOLE)
sh.setFormatter(console_formatter)

# Filter out redirected output on console, to avoid duplicates
filt = StreamFilter()
sh.addFilter(filt)

log.addHandler(sh)


def reroute_output():
    """Route stdout and stderr to logger (custom handler)"""
    if (getattr(sys, 'frozen', False)
       or sys.stdout != sys.__stdout__):
        return
    sys.stdout = StreamToLogger(sys.stdout, logging.INFO)
    sys.stderr = StreamToLogger(sys.stderr, logging.WARNING)


def set_level_file(level=logging.INFO):
    """Adjust the minimum log level written to our logfile"""
    fh.setLevel(level)


def set_level_console(level=logging.INFO):
    """Adjust the minimum log level for output to the terminal"""
    sh.setLevel(level)



"""
 @file
 @brief This file contains the tutorial dialogs, which are used to explain certain features to new users
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import functools

from PyQt5.QtCore import Qt, QPoint, QRectF, QTimer, QObject, QRect
from PyQt5.QtGui import (
    QColor, QPalette, QPen, QPainter, QPainterPath, QKeySequence,
)
from PyQt5.QtWidgets import (
    QAction, QLabel, QWidget, QVBoxLayout, QHBoxLayout,
    QPushButton, QToolButton, QCheckBox,
)
from classes.logger import log
from classes.app import get_app
from classes.metrics import track_metric_screen
from classes import sentry


class TutorialDialog(QWidget):
    """ A customized QWidget used to instruct a user how to use a certain feature """

    def paintEvent(self, event):
        """ Custom paint event """
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        # Set correct margins based on left/right arrow
        arrow_width = 15
        if not self.draw_arrow_on_right:
            self.vbox.setContentsMargins(45, 10, 20, 10)
        else:
            self.vbox.setContentsMargins(20, 10, 45, 10)

        # Define rounded rectangle geometry
        corner_radius = 10
        if self.draw_arrow_on_right:
            # Rectangle starts at left edge; arrow is on the right
            rounded_rect = QRectF(0, 0, self.width() - arrow_width, self.height())
        else:
            # Rectangle shifted to the right; arrow is on the left
            rounded_rect = QRectF(arrow_width, 0, self.width() - arrow_width, self.height())

        # Clip to the rounded rectangle
        path = QPainterPath()
        path.addRoundedRect(rounded_rect, corner_radius, corner_radius)
        painter.setClipPath(path)

        # Fill background
        frameColor = QColor("#53a0ed")
        painter.setPen(QPen(frameColor, 1.2))
        painter.setBrush(self.palette().color(QPalette.Window))
        painter.drawRoundedRect(rounded_rect, corner_radius, corner_radius)

        # Disable clipping temporarily for the arrow
        painter.setClipping(False)

        # Draw arrow if needed
        if self.arrow:
            arrow_height = 15
            arrow_offset = 35

            if self.draw_arrow_on_right:
                # Arrow on the right side
                arrow_point = rounded_rect.topRight().toPoint() + QPoint(arrow_width, arrow_offset)
                arrow_top_corner = rounded_rect.topRight().toPoint() + QPoint(-1, arrow_offset - arrow_height)
                arrow_bottom_corner = rounded_rect.topRight().toPoint() + QPoint(-1, arrow_offset + arrow_height)
            else:
                # Arrow on the left side
                arrow_point = rounded_rect.topLeft().toPoint() + QPoint(-arrow_width, arrow_offset)
                arrow_top_corner = rounded_rect.topLeft().toPoint() + QPoint(1, arrow_offset - arrow_height)
                arrow_bottom_corner = rounded_rect.topLeft().toPoint() + QPoint(1, arrow_offset + arrow_height)

            # Draw triangle (filled with the same background color as the window)
            path = QPainterPath()
            path.moveTo(arrow_point)  # Arrow tip
            path.lineTo(arrow_top_corner)  # Top corner of the triangle
            path.lineTo(arrow_bottom_corner)  # Bottom corner of the triangle
            path.closeSubpath()
            painter.fillPath(path, self.palette().color(QPalette.Window))

            # Draw the triangle's borders
            border_pen = QPen(frameColor, 1)
            painter.setPen(border_pen)
            painter.drawLine(arrow_point, arrow_top_corner)  # Top triangle border
            painter.drawLine(arrow_point, arrow_bottom_corner)  # Bottom triangle border

    def checkbox_metrics_callback(self, state):
        """ Callback for error and anonymous usage checkbox"""
        s = get_app().get_settings()
        if state == Qt.Checked:
            # Enabling metrics sending
            s.set("send_metrics", True)
            sentry.init_tracing()

            # Opt-in for metrics tracking
            track_metric_screen("metrics-opt-in")
        else:
            # Opt-out for metrics tracking
            track_metric_screen("metrics-opt-out")
            sentry.disable_tracing()

            # Disable metric sending
            s.set("send_metrics", False)

    def mouseReleaseEvent(self, event):
        """Process click events on tutorial. Especially useful when tutorial messages are partially
        obscured or offscreen (i.e. on small screens). Just click any part of the tutorial, and it will
        move on to the next one."""
        self.manager.next_tip(self.widget_id)

    def __init__(self, widget_id, text, arrow, manager, *args):
        super().__init__(*args)

        # Ensure frameless, floating behavior
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.Tool | Qt.WindowStaysOnTopHint)
        self.setAttribute(Qt.WA_NoSystemBackground, True)
        self.setAttribute(Qt.WA_TranslucentBackground, True)
        self.setAttribute(Qt.WA_DeleteOnClose, True)

        # get translations
        app = get_app()
        _ = app._tr

        # Keep track of widget to position next to
        self.widget_id = widget_id
        self.arrow = arrow
        self.manager = manager
        self.draw_arrow_on_right = False

        # Create vertical box
        self.vbox = QVBoxLayout()

        # Add label
        self.label = QLabel(self)
        self.label.setObjectName("lblTutorialText")
        self.label.setText(text)
        self.label.setTextFormat(Qt.RichText)
        self.label.setWordWrap(True)
        self.label.setStyleSheet("")
        self.label.setAttribute(Qt.WA_TransparentForMouseEvents)
        self.vbox.addWidget(self.label)

        # Add error and anonymous metrics checkbox (for ID=0) tooltip
        # This is a bit of a hack, but since it's the only exception, it's
        # probably okay for now.
        if self.widget_id == "0":
            # Get settings
            s = get_app().get_settings()

            # create spinner
            checkbox_metrics = QCheckBox()
            checkbox_metrics.setObjectName("checkboxMetrics")
            checkbox_metrics.setText(_("Yes, I would like to improve OpenShot!"))
            if s.get("send_metrics"):
                checkbox_metrics.setCheckState(Qt.Checked)
            else:
                checkbox_metrics.setCheckState(Qt.Unchecked)
            checkbox_metrics.stateChanged.connect(functools.partial(self.checkbox_metrics_callback))
            self.vbox.addWidget(checkbox_metrics)

        # Add button box
        hbox = QHBoxLayout()
        hbox.setContentsMargins(0, 5, 0, 5)

        # Close action
        self.close_action = QAction(_("Hide Tutorial"), self)
        self.close_action.setShortcut(QKeySequence(Qt.Key_Escape))
        self.close_action.setShortcutContext(Qt.ApplicationShortcut)

        # Create buttons
        self.btn_close_tips = QPushButton(self)
        self.btn_close_tips.setText(_("Hide Tutorial"))
        self.btn_close_tips.setObjectName("HideTutorial")
        self.btn_close_tips.addAction(self.close_action)

        self.btn_next_tip = QPushButton(self)
        self.btn_next_tip.setObjectName("NextTip")
        self.btn_next_tip.setText(_("Next"))
        self.btn_next_tip.setStyleSheet("font-weight:bold;")

        hbox.addWidget(self.btn_close_tips)
        hbox.addWidget(self.btn_next_tip)
        self.vbox.addLayout(hbox)

        # Set layout, cursor, and size
        self.setLayout(self.vbox)
        self.setCursor(Qt.ArrowCursor)
        self.setMinimumWidth(350)
        self.setMinimumHeight(100)
        self.setFocusPolicy(Qt.ClickFocus)

        # Connect close action signal
        self.close_action.triggered.connect(
            functools.partial(self.manager.hide_tips, self.widget_id, True))


class TutorialManager(QObject):
    """ Manage and present a list of tutorial dialogs """

    def process(self):
        """ Process and show the first non-completed tutorial """

        # If a tutorial is already visible, just update it
        if self.current_dialog:
            # Respond to possible dock floats/moves
            self.dock.raise_()
            self.re_position_dialog()
            return

        # Loop through and add each tutorial dialog
        for tutorial_details in self.tutorial_objects:
            # Get details
            tutorial_id = tutorial_details["id"]

            # Get QWidget
            tutorial_object = self.get_object(tutorial_details["object_id"])

            # Skip completed tutorials (and invisible widgets)
            if not self.tutorial_enabled or tutorial_id in self.tutorial_ids or tutorial_object.visibleRegion().isEmpty():
                continue

            # Create tutorial
            self.position_widget = tutorial_object
            self.offset = QPoint(
                int(tutorial_details["x"]),
                int(tutorial_details["y"]))
            tutorial_dialog = TutorialDialog(tutorial_id, tutorial_details["text"], tutorial_details["arrow"], self)
            tutorial_dialog.setObjectName("tutorial")

            # Connect signals
            tutorial_dialog.btn_next_tip.clicked.connect(functools.partial(self.next_tip, tutorial_id))
            tutorial_dialog.btn_close_tips.clicked.connect(functools.partial(self.hide_tips, tutorial_id, True))

            # Get previous dock contents
            old_widget = self.dock.widget()

            # Insert into tutorial dock
            self.dock.setWidget(tutorial_dialog)
            self.current_dialog = tutorial_dialog

            # Show dialog
            self.dock.adjustSize()
            self.dock.setEnabled(True)
            self.re_position_dialog()
            self.dock.show()

            # Delete old widget
            if old_widget:
                old_widget.close()

            break

    def get_object(self, object_id):
        """Get an object from the main window by object id"""
        if object_id == "filesView":
            return self.win.filesView
        elif object_id == "timeline":
            return self.win.timeline
        elif object_id == "dockVideo":
            return self.win.dockVideo
        elif object_id == "propertyTableView":
            return self.win.propertyTableView
        elif object_id == "transitionsView":
            return self.win.transitionsView
        elif object_id == "effectsView":
            return self.win.effectsView
        elif object_id == "emojisView":
            return self.win.emojiListView
        elif object_id == "actionPlay":
            # Find play/pause button on transport controls toolbar
            for w in self.win.actionPlay.associatedWidgets():
                if isinstance(w, QToolButton) and w.isVisible():
                    return w
        elif object_id == "export_button":
            # Find export toolbar button on main window
            for w in reversed(self.win.actionExportVideo.associatedWidgets()):
                if isinstance(w, QToolButton) and  w.isVisible() and w.parent() == self.win.toolBar:
                    return w

    def next_tip(self, tid):
        """ Mark the current tip completed, and show the next one """
        # Hide matching tutorial
        self.hide_tips(tid)

        # Advance to the next one
        self.process()

    def hide_tips(self, tid, user_clicked=False):
        """ Hide the current tip, and don't show anymore """
        s = get_app().get_settings()

        # Loop through and find current tid
        for tutorial_object in self.tutorial_objects:
            # Get details
            tutorial_id = tutorial_object["id"]
            if tutorial_id == tid:
                # Hide dialog
                self.close_dialogs()
                # Update settings that this tutorial is completed
                if tid not in self.tutorial_ids:
                    self.tutorial_ids.append(str(tid))
                    s.set("tutorial_ids", ",".join(self.tutorial_ids))

        # Mark tutorial as completed (if settings)
        if user_clicked:
            # Disable all tutorials
            self.tutorial_enabled = False
            s.set("tutorial_enabled", False)

        # Forgot current tutorial
        self.current_dialog = None

    def close_dialogs(self):
        """ Close any open tutorial dialogs """
        if self.current_dialog:
            self.dock.hide()
            self.dock.setEnabled(False)
            self.current_dialog = None

    def exit_manager(self):
        """ Disconnect from all signals, and shutdown tutorial manager """
        try:
            self.win.dockFiles.visibilityChanged.disconnect()
            self.win.dockTransitions.visibilityChanged.disconnect()
            self.win.dockEffects.visibilityChanged.disconnect()
            self.win.dockProperties.visibilityChanged.disconnect()
            self.win.dockVideo.visibilityChanged.disconnect()
            self.tutorial_timer.timeout.disconnect()
        except Exception:
            log.debug('Failed to properly disconnect from dock signals', exc_info=1)

        # Close dialog window
        self.close_dialogs()

    def re_show_dialog(self):
        """ Re show an active dialog """
        if self.current_dialog:
            self.dock.update()
            self.dock.raise_()
            self.dock.show()

    def hide_dialog(self):
        """ Hide an active dialog """
        if self.current_dialog:
            self.dock.hide()

    def re_position_dialog(self):
        """ Reposition the tutorial dialog next to self.position_widget. """
        # Bail if no dialog or target widget hidden
        if not self.current_dialog:
            return
        if self.position_widget.isHidden() or self.position_widget.visibleRegion().isEmpty():
            self.hide_dialog()
            return

        # Compute the reference rect of the target widget
        pos_rect = self.position_widget.rect()
        # “float” the popup 1/4 size away from top-left corner
        pos_rect.setSize(pos_rect.size() / 4)
        pos_rect.translate(self.offset)

        # Compute both possible positions (arrow on left vs. arrow on right)
        # NOTE: We do this BEFORE we actually move the dialog!
        position_arrow_left = self.position_widget.mapToGlobal(pos_rect.bottomRight())
        position_arrow_right = self.position_widget.mapToGlobal(pos_rect.bottomLeft()) - QPoint(
            self.current_dialog.width(), 0)

        # Decide which side is viable. For example, we can see if arrow-on-left
        # would run off the right side of the screen. If it does, pick arrow-on-right.
        screen_rect = get_app().primaryScreen().availableGeometry()
        monitor_width = screen_rect.width()

        # If placing “arrow on left” means we’d exceed monitor width, we must switch to arrow on right
        would_exceed_right_edge = (position_arrow_left.x() + self.current_dialog.width()) > monitor_width
        if would_exceed_right_edge:
            final_position = position_arrow_right
            arrow_on_right = True
        else:
            final_position = position_arrow_left
            arrow_on_right = False

        # Update the dialog’s internal state (so paintEvent() knows how to draw it).
        self.current_dialog.draw_arrow_on_right = arrow_on_right

        # Update margins ONE time here, so geometry only changes once
        if arrow_on_right:
            self.current_dialog.vbox.setContentsMargins(20, 10, 45, 10)
        else:
            self.current_dialog.vbox.setContentsMargins(45, 10, 20, 10)

        # Move the dock exactly once, and raise it
        self.dock.move(final_position)
        self.re_show_dialog()

    def process_visibility(self):
        """Handle callbacks when widget visibility changes"""
        if self.tutorial_enabled:
            self.tutorial_timer.start()

    def __init__(self, win, *args):
        # Init QObject superclass
        super().__init__(*args)

        """ Constructor """
        self.win = win
        self.dock = win.dockTutorial
        self.current_dialog = None
        self.dock.setParent(None)

        # get translations
        app = get_app()
        _ = app._tr

        # get settings
        s = app.get_settings()
        self.tutorial_enabled = s.get("tutorial_enabled")
        self.tutorial_ids = s.get("tutorial_ids").split(",")

        # Add all possible tutorials
        self.tutorial_objects = [
            {"id": "0",
             "x": 0,
             "y": 0,
             "object_id": "dockVideo",
             "text": _("<b>Welcome!</b> OpenShot Video Editor is an award-winning, open-source video editing application! This tutorial will walk you through the basics.<br><br>Would you like to automatically send errors and metrics to help improve OpenShot?"),
             "arrow": False
             },
            {"id": "1",
             "x": 0,
             "y": 0,
             "object_id": "filesView",
             "text": _("<b>Project Files:</b> Get started with your project by adding video, audio, and image files here. Drag and drop files from your file system."),
             "arrow": True
             },
            {"id": "2",
             "x": 0,
             "y": 0,
             "object_id": "timeline",
             "text": _("<b>Timeline:</b> Arrange your clips on the timeline here. Overlap clips to create automatic transitions. Access lots of fun presets and options by right-clicking on clips."),
             "arrow": True
             },
            {"id": "3",
             "x": 10,
             "y": -42,
             "object_id": "actionPlay",
             "text": _("<b>Video Preview:</b> Watch your timeline video preview here. Use the buttons (play, rewind, fast-forward) to control the video playback."),
             "arrow": True},
            {"id": "4",
             "x": 0,
             "y": 0,
             "object_id": "propertyTableView",
             "text": _("<b>Properties:</b> View and change advanced properties of clips and effects here. Right-clicking on clips is usually faster than manually changing properties."),
             "arrow": True
             },
            {"id": "5",
             "x": 0,
             "y": 0,
             "object_id": "transitionsView",
             "text": _("<b>Transitions:</b> Create a gradual fade from one clip to another. Drag and drop a transition onto the timeline and position it on top of a clip (usually at the beginning or ending)."),
             "arrow": True
             },
            {"id": "6",
             "x": 0,
             "y": 0,
             "object_id": "effectsView",
             "text": _("<b>Effects:</b> Adjust brightness, contrast, saturation, and add exciting special effects. Drag and drop an effect onto the timeline and position it on top of a clip (or track)"),
             "arrow": True
             },
            {"id": "8",
             "x": 0,
             "y": 0,
             "object_id": "emojisView",
             "text": _("<b>Emojis:</b> Add exciting and colorful emojis to your project! Drag and drop an emoji onto the timeline. The emoji will become a new Clip when dropped on the Timeline."),
             "arrow": True
             },
            {"id": "7",
             "x": 10,
             "y": -27,
             "object_id": "export_button",
             "text": _("<b>Export Video:</b> When you are ready to create your finished video, click this button to export your timeline as a single video file."),
             "arrow": True
             }
        ]

        # Configure tutorial frame
        self.dock.setTitleBarWidget(QWidget())  # Prevents window decoration
        self.dock.setAttribute(Qt.WA_NoSystemBackground, True)
        self.dock.setAttribute(Qt.WA_TranslucentBackground, True)
        self.dock.setWindowFlags(Qt.FramelessWindowHint | Qt.Tool | Qt.WindowStaysOnTopHint)
        self.dock.setFloating(True)

        # Timer for processing new tutorials
        self.tutorial_timer = QTimer(self)
        self.tutorial_timer.setInterval(200)
        self.tutorial_timer.setSingleShot(True)
        self.tutorial_timer.timeout.connect(self.process)

        # Connect to interface dock widgets
        self.win.dockFiles.visibilityChanged.connect(self.process_visibility)
        self.win.dockTransitions.visibilityChanged.connect(self.process_visibility)
        self.win.dockEffects.visibilityChanged.connect(self.process_visibility)
        self.win.dockProperties.visibilityChanged.connect(self.process_visibility)
        self.win.dockVideo.visibilityChanged.connect(self.process_visibility)
        self.win.dockEmojis.visibilityChanged.connect(self.process_visibility)



"""
 @file
 @brief Clip and transition interaction helpers for the timeline widget.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import json
import time
import uuid
from PyQt5.QtCore import Qt, QRectF
from PyQt5.QtWidgets import QApplication
from classes.app import get_app
from classes.clip_utils import is_single_image_media
from classes.query import Clip, Transition
from classes.waveform import SAMPLES_PER_SECOND as WAVEFORM_SAMPLES_PER_SECOND
from windows.views.retime import retime_clip


class ClipInteractionMixin:
    def _selected_resize_candidates(self):
        """Return selected clips/transitions from the current geometry."""
        items = []
        geometry = getattr(self, "geometry", None)
        if geometry is None or not hasattr(geometry, "iter_items"):
            return items
        for _rect, item, selected, _type in geometry.iter_items(viewport=False):
            if selected:
                items.append(item)
        return items

    def _resize_edge_value(self, item, edge):
        """Return the timeline position of the requested item's edge."""
        data = getattr(item, "data", None)
        if not isinstance(data, dict):
            return None
        try:
            position = float(data.get("position", 0.0) or 0.0)
            start = float(data.get("start", 0.0) or 0.0)
            end = float(data.get("end", start) or start)
        except (TypeError, ValueError):
            return None
        if edge == "left":
            return position
        if edge == "right":
            return position + max(0.0, end - start)
        return None

    def _resize_edge_tolerance(self):
        fps = self._positive_float(getattr(self, "fps_float", None))
        if fps:
            # Allow one-frame drift so mixed clip/transition selections keep
            # behaving as a shared edge after commit quantization.
            return max(1e-6, (1.0 / fps) + 1e-9)
        return 1e-6

    def _resize_targets_for_item(self, item, edge):
        """
        Resolve which selected items should respond to an edge resize.

        For multi-selection, only shared outer edges are resizable. Interior
        edges fall back to normal drag behavior.
        """
        if edge not in ("left", "right") or item is None:
            return []

        item_id = str(getattr(item, "id", "") or "")
        selected_items = self._selected_resize_candidates()
        selected_ids = {
            str(getattr(candidate, "id", "") or "")
            for candidate in selected_items
        }

        if item_id not in selected_ids or len(selected_items) <= 1:
            return [item]

        candidate_edge = self._resize_edge_value(item, edge)
        if candidate_edge is None:
            return []

        edge_values = []
        for candidate in selected_items:
            value = self._resize_edge_value(candidate, edge)
            if value is not None:
                edge_values.append(value)
        if not edge_values:
            return []

        outer_edge = min(edge_values) if edge == "left" else max(edge_values)
        tolerance = self._resize_edge_tolerance()
        if abs(candidate_edge - outer_edge) > tolerance:
            return []

        return [
            candidate
            for candidate in selected_items
            if abs((self._resize_edge_value(candidate, edge) or 0.0) - outer_edge) <= tolerance
        ]

    def _build_resize_context(self, item):
        """Capture the per-item baseline state used during trim/retime preview."""
        world_rect = self.geometry.calc_item_rect(item)
        rect = self.geometry.calc_item_rect(item, viewport=True)
        initial = {
            "start": float(item.data.get("start", 0.0)),
            "end": float(item.data.get("end", 0.0)),
            "position": float(item.data.get("position", 0.0)),
            "duration": float(item.data.get("duration", item.data.get("end", 0.0) - item.data.get("start", 0.0))),
        }
        context = {
            "item": item,
            "world_rect": QRectF(world_rect),
            "rect": QRectF(rect),
            "initial": initial,
            "max_duration": None,
            "clip_is_single_image": False,
            "allow_left_overflow": False,
            "static_mask": False,
        }
        if isinstance(item, Clip):
            max_duration = self._clip_reader_duration_seconds(item)
            if max_duration is None:
                max_duration = self._positive_float(initial.get("duration"))
            current_end = initial["end"]
            if max_duration is not None and current_end > max_duration:
                max_duration = current_end
            clip_is_single_image = self._clip_is_single_image(item)
            context["max_duration"] = max_duration
            context["clip_is_single_image"] = clip_is_single_image
            context["allow_left_overflow"] = bool(self.enable_timing or clip_is_single_image)
        else:
            context["static_mask"] = self._transition_uses_static_mask(item.data)
        return context

    def _clear_resize_preview_overrides(self, item):
        if isinstance(item, Transition):
            self._pending_transition_overrides.pop(item.id, None)
        else:
            self._pending_clip_overrides.pop(item.id, None)

    def _apply_resize_preview_override(self, item, context, start, end, position):
        if isinstance(item, Clip):
            override = self._pending_clip_overrides.setdefault(
                item.id,
                {
                    "start": start,
                    "end": end,
                    "position": position,
                    "initial_start": context["initial"].get("start", start),
                    "initial_end": context["initial"].get("end", end),
                    "initial_position": context["initial"].get("position", position),
                },
            )
            override["start"] = start
            override["end"] = end
            override["position"] = position
            override["scale"] = bool(self.enable_timing)
        else:
            override = self._pending_transition_overrides.setdefault(
                item.id,
                {
                    "position": position,
                    "start": start,
                    "end": end,
                    "initial_start": context["initial"].get("start", start),
                    "initial_end": context["initial"].get("end", end),
                },
            )
            override["position"] = position
            override["start"] = start
            override["end"] = end
            override["scale"] = bool(context.get("static_mask"))

    def _restore_resize_snap_ignore_ids(self, resized_items):
        if hasattr(self, "_resize_snap_ignore_backup"):
            ignore_ids = set(self._resize_snap_ignore_backup)
            del self._resize_snap_ignore_backup
        else:
            ignore_ids = set(getattr(self, "_snap_ignore_ids", set()))

        resized_ids = {
            getattr(item, "id", None)
            for item in (resized_items or [])
        }
        self._snap_ignore_ids = {
            item_id for item_id in ignore_ids if item_id not in resized_ids
        }

    def _finalize_resize_preview_state(self, resize_items, backend_refresh_requested):
        """
        End a resize gesture's preview state.

        Resize commits are now batched, so once the saves complete we can drop
        all preview overrides immediately and rebuild from committed data.
        """
        for candidate in resize_items:
            self._clear_resize_preview_overrides(candidate)
        self.changed(None)

    def _normalize_resize_commit_bounds(self, start, end, position):
        """
        Quantize resize results while preserving the snapped outer right edge.

        Snapping ``position``, ``start``, and ``end`` independently can move
        the item's timeline right edge by a frame, which breaks later
        multi-selection shared-edge detection. Snap the absolute right edge
        once, then derive ``end`` from the snapped left-side values.
        """
        snapped_position = self._snap_time(position)
        snapped_start = self._snap_time(start)
        snapped_right = self._snap_time(position + max(0.0, end - start))

        fps = float(getattr(self, "fps_float", 0.0) or 0.0)
        min_len = (1.0 / fps) if fps > 0.0 else 0.0
        snapped_span = max(min_len, snapped_right - snapped_position)
        snapped_end = self._snap_time(snapped_start + snapped_span)
        if min_len > 0.0 and snapped_end - snapped_start < min_len:
            snapped_end = snapped_start + min_len
        return snapped_start, snapped_end, snapped_position

    def _commit_resized_clip(self, clip, start, end, position, context, transaction_id, ignore_refresh):
        """Persist a resized clip, batching retime commits the same way as trims."""
        if self.enable_timing:
            initial_start = context.get("initial", {}).get("start")
            if initial_start is None:
                initial_start = start
            original_start = float(initial_start)
            duration = end - start
            target_end = self._snap_time(original_start + duration)
            target_position = self._snap_time(position)
            if not retime_clip(clip, target_end, target_position, direction=1):
                return False
            ui_data = clip.data.get("ui")
            if isinstance(ui_data, dict) and "audio_data" in ui_data:
                ui_data.pop("audio_data", None)
            self.update_clip_data(
                clip.data,
                only_basic_props=False,
                ignore_reader=True,
                ignore_refresh=ignore_refresh,
                transaction_id=transaction_id,
            )
            return True

        start, end, position = self._normalize_resize_commit_bounds(start, end, position)
        clip.data["start"] = start
        clip.data["end"] = end
        clip.data["position"] = position
        self.update_clip_data(
            clip.data,
            only_basic_props=True,
            ignore_reader=True,
            ignore_refresh=ignore_refresh,
            transaction_id=transaction_id,
        )
        return True

    def _active_resize_items(self):
        """Return the items currently participating in an edge resize gesture."""
        if getattr(self, "_press_hit", "") != "clip-edge":
            return []
        items = list(getattr(self, "_resize_items", None) or [])
        if items:
            return items
        item = getattr(self, "_resizing_item", None)
        return [item] if item is not None else []

    def _is_active_resize_item(self, item):
        """Return True when *item* is part of the current edge-resize preview."""
        if item is None:
            return False
        for candidate in self._active_resize_items():
            if candidate is item:
                return True
            if getattr(candidate, "id", None) == getattr(item, "id", None):
                return True
        return False

    def _resize_preview_focus_item(self):
        """Choose the primary grabbed item for player preview updates."""
        items = list(self._active_resize_items())
        if not items:
            return None
        primary = getattr(self, "_resizing_item", None)
        if primary is not None:
            primary_id = getattr(primary, "id", None)
            for item in items:
                if item is primary or getattr(item, "id", None) == primary_id:
                    return item
        return items[0]

    def _transition_uses_static_mask(self, transition_data):
        reader = {}
        if isinstance(transition_data, dict):
            for key in ("mask_reader", "reader"):
                candidate = transition_data.get(key)
                if isinstance(candidate, dict):
                    reader = candidate
                    break
        if isinstance(reader, dict) and "has_single_image" in reader:
            return bool(reader.get("has_single_image"))
        return bool(is_single_image_media(reader))

    def _set_trim_thumbnail_suspension(self, enabled, clip_id=None):
        """Pause thumbnail generation while trimming and drop stale queued work."""
        self._suspend_thumbnail_requests = bool(enabled)
        clip_key = str(clip_id or "")
        if enabled:
            if self.thumbnail_manager:
                self.thumbnail_manager.clear_pending()
            if clip_key and hasattr(self, "clip_painter"):
                # Keep existing cached thumbs visible while trimming, but drop in-flight requests.
                self.clip_painter.invalidate_clip_thumbnails(
                    clip_key,
                    drop_cache=False,
                    drop_pending=True,
                    drop_fallback=False,
                    drop_preview=False,
                    invalidate_render_cache=False,
                )
            return

        if self.thumbnail_manager:
            self.thumbnail_manager.clear_pending()
        if clip_key and hasattr(self, "clip_painter"):
            # Preserve existing thumbnails until replacements arrive; only drop stale in-flight requests.
            self.clip_painter.invalidate_clip_thumbnails(
                clip_key,
                drop_cache=False,
                drop_pending=True,
                drop_fallback=False,
                invalidate_render_cache=False,
            )
            self.clip_painter.clear_retime_preview(clip_key)

    def clip_has_pending_override(self, clip):
        if not isinstance(clip, Clip):
            return False
        return clip.id in self._pending_clip_overrides

    def clip_waveform_window(self, clip):
        data = clip.data if isinstance(clip.data, dict) else {}
        start = float(data.get("start", 0.0) or 0.0)
        end = float(data.get("end", start) or start)
        if end < start:
            end = start
        overrides = None
        if isinstance(clip, Clip):
            overrides = self._pending_clip_overrides.get(clip.id)

        pending_start = start
        pending_end = end
        initial_start = start
        initial_end = end
        scale_waveform = False
        if overrides:
            pending_start = float(overrides.get("start", pending_start) or pending_start)
            pending_end = float(overrides.get("end", pending_end) or pending_end)
            initial_start = float(overrides.get("initial_start", initial_start) or initial_start)
            initial_end = float(overrides.get("initial_end", initial_end) or initial_end)
            if pending_end < pending_start:
                pending_end = pending_start
            if initial_end < initial_start:
                initial_end = initial_start
            scale_waveform = bool(overrides.get("scale"))

        samples_per_second = getattr(self, "_waveform_samples_per_second", None)
        if not samples_per_second:
            try:
                samples_per_second = int(WAVEFORM_SAMPLES_PER_SECOND)
            except Exception:
                samples_per_second = 20
            if samples_per_second <= 0:
                samples_per_second = 20
            self._waveform_samples_per_second = samples_per_second

        ui_data = data.get("ui", {}) if isinstance(data, dict) else {}
        audio_data = ui_data.get("audio_data") if isinstance(ui_data, dict) else None
        sample_count = len(audio_data) if isinstance(audio_data, list) else 0
        media_duration = 0.0
        if sample_count:
            media_duration = float(sample_count) / float(samples_per_second)

        if media_duration <= 0.0:
            media_duration = max(initial_end, pending_end, end, start, 0.0)

        clip_span = max(initial_end - initial_start, 0.0)
        tolerance = 1.0 / float(samples_per_second)
        dataset_matches_clip = (
            media_duration > 0.0
            and clip_span > 0.0
            and abs(media_duration - clip_span) <= max(tolerance, clip_span * 1e-3)
        )
        origin = initial_start if dataset_matches_clip else 0.0

        def _ratio(value, offset):
            if media_duration <= 0.0:
                return 0.0
            relative = float(value) - float(offset)
            if relative < 0.0:
                relative = 0.0
            if relative > media_duration:
                relative = media_duration
            return relative / media_duration

        start_ratio = _ratio(pending_start, origin)
        end_ratio = _ratio(pending_end, origin)
        source_start_ratio = _ratio(initial_start, origin)
        source_end_ratio = _ratio(initial_end, origin)

        if end_ratio < start_ratio:
            end_ratio = start_ratio
        if source_end_ratio < source_start_ratio:
            source_end_ratio = source_start_ratio

        return {
            "start_ratio": start_ratio,
            "end_ratio": end_ratio,
            "scale": scale_waveform,
            "source_start_ratio": source_start_ratio,
            "source_end_ratio": source_end_ratio,
        }

    def clip_waveform_cache_token(self, clip):
        data = clip.data if isinstance(clip.data, dict) else {}
        ui_data = data.get("ui", {}) if isinstance(data, dict) else {}
        audio_data = ui_data.get("audio_data") if isinstance(ui_data, dict) else None
        if isinstance(audio_data, list):
            return (
                len(audio_data),
                ui_data.get("waveform_token"),
            )
        return 0

    def _clip_data_dict(self, clip):
        data = getattr(clip, "data", None)
        return data if isinstance(data, dict) else {}

    def _clip_reader_dict(self, clip):
        reader = self._clip_data_dict(clip).get("reader")
        return reader if isinstance(reader, dict) else {}

    def _clip_is_single_image(self, clip):
        data = self._clip_data_dict(clip)
        if is_single_image_media(data):
            return True
        reader = self._clip_reader_dict(clip)
        if is_single_image_media(reader):
            return True
        return False

    @staticmethod
    def _float_or_none(value):
        try:
            return float(value)
        except (TypeError, ValueError):
            return None

    def _positive_float(self, value):
        parsed = self._float_or_none(value)
        if parsed is None or parsed <= 0.0:
            return None
        return parsed

    def _clip_reader_duration_seconds(self, clip):
        data = self._clip_data_dict(clip)
        reader = self._clip_reader_dict(clip)
        start = self._float_or_none(data.get("start"))
        if start is None:
            start = 0.0

        candidate_limits = []

        duration = self._positive_float(reader.get("duration"))
        if duration is not None:
            candidate_limits.append(duration)

        video_length = self._positive_float(reader.get("video_length"))
        fps_meta = reader.get("fps") if isinstance(reader.get("fps"), dict) else {}
        fps_num = self._positive_float(fps_meta.get("num"))
        fps_den = self._positive_float(fps_meta.get("den"))
        if video_length is not None and fps_num is not None and fps_den is not None and fps_den > 0.0:
            fps_value = fps_num / fps_den
            if fps_value > 0.0:
                candidate_limits.append(video_length / fps_value)

        project_fps = self._positive_float(getattr(self, "fps_float", None))
        time_data = data.get("time") if isinstance(data.get("time"), dict) else {}
        time_points = time_data.get("Points") if isinstance(time_data.get("Points"), list) else []
        if project_fps is not None and len(time_points) >= 2:
            x_values = []
            for point in time_points:
                if not isinstance(point, dict):
                    continue
                co = point.get("co")
                if not isinstance(co, dict):
                    continue
                x_val = self._float_or_none(co.get("X"))
                if x_val is not None:
                    x_values.append(x_val)
            if len(x_values) >= 2:
                time_duration = (max(x_values) - min(x_values)) / project_fps
                if time_duration > 0.0:
                    candidate_limits.append(start + time_duration)

        clip_duration = self._positive_float(data.get("duration"))
        if clip_duration is not None:
            candidate_limits.append(start + clip_duration)

        end = self._float_or_none(data.get("end"))
        if end is None:
            end = start
        span = end - start
        if span > 0.0:
            candidate_limits.append(end)

        return max(candidate_limits) if candidate_limits else None

    def _selection_anchor_from_item(self, item, sel_type=None):
        """Build a normalized selection anchor for clips/transitions."""
        data = item.data if isinstance(item.data, dict) else {}
        item_type = sel_type or ("transition" if isinstance(item, Transition) else "clip")
        position = float(data.get("position", 0.0) or 0.0)
        layer = int(data.get("layer", 0) or 0)
        start = float(data.get("start", 0.0) or 0.0)
        end = float(data.get("end", start) or start)
        duration = max(0.0, end - start)
        return {
            "type": item_type,
            "position": position,
            "end_position": position + duration,
            "layer": layer,
        }

    def _normalize_selection_anchor(self, anchor):
        """Convert legacy tuple anchors into the normalized anchor dict format."""
        if isinstance(anchor, dict):
            try:
                position = float(anchor.get("position", 0.0) or 0.0)
                end_position = float(anchor.get("end_position", position) or position)
                layer = int(anchor.get("layer", 0) or 0)
            except (TypeError, ValueError):
                return None
            if end_position < position:
                end_position = position
            return {
                "type": anchor.get("type"),
                "position": position,
                "end_position": end_position,
                "layer": layer,
            }

        if isinstance(anchor, (tuple, list)) and len(anchor) >= 2:
            try:
                position = float(anchor[0] or 0.0)
                layer = int(anchor[1] or 0)
            except (TypeError, ValueError):
                return None
            return {
                "type": None,
                "position": position,
                "end_position": position,
                "layer": layer,
            }

        return None

    def _startClipDrag(self):
        """Begin a drag operation on one or many selected clips/transitions."""
        e = self._last_event

        self.snap.reset()
        self._collapse_selection_on_release = False
        self._collapse_selection_target = None
        self._drag_moved = False
        self._drag_press_pos = e.pos() if e else None
        self._drag_threshold_met = False

        # Identify the item under the cursor (include clips and transitions)
        clicked_item = None
        for rect, item, _selected, _type in self.geometry.iter_items(reverse=True):
            if rect.contains(e.pos()):
                clicked_item = item
                break
        if clicked_item is None:
            return

        self._fix_cursor(self.cursors["hand"])

        # Each drag operation is grouped under a single undo transaction
        self._drag_transaction_id = str(uuid.uuid4())

        modifiers = e.modifiers()
        ctrl = bool(modifiers & Qt.ControlModifier)
        alt = bool(modifiers & Qt.AltModifier)
        shift = bool(modifiers & Qt.ShiftModifier)
        sel_type = "transition" if isinstance(clicked_item, Transition) else "clip"
        last_anchor = self._normalize_selection_anchor(getattr(self, "_last_click_pos", None))
        already = (
            clicked_item.id in self.win.selected_clips or
            clicked_item.id in self.win.selected_transitions
        )

        if alt:
            # ALT+Click: select the clicked item plus everything to its right on
            # the same layer.  CTRL+ALT keeps existing selection; plain ALT replaces.
            if not ctrl:
                self.win.clearSelections()
            self.selectRipple(clicked_item.id, sel_type)
            # ALT does not update _last_click_pos; nothing to drag.
            self.dragging_items = []
            self._drag_transaction_id = None
            self.changed(None)
            return

        elif shift and last_anchor:
            # SHIFT+Click: simple rectangular range-select between anchor and
            # click (position + layer), across both clips and transitions.
            current_anchor = self._selection_anchor_from_item(clicked_item, sel_type)
            min_pos = min(last_anchor["position"], current_anchor["position"])
            max_pos = max(last_anchor["position"], current_anchor["position"])
            min_layer = min(last_anchor["layer"], current_anchor["layer"])
            max_layer = max(last_anchor["layer"], current_anchor["layer"])
            eps = 1e-9
            matched_items = []
            for item in Clip.filter() + Transition.filter():
                d = item.data if isinstance(item.data, dict) else {}
                try:
                    item_pos = float(d.get("position", 0.0) or 0.0)
                    item_layer = int(d.get("layer", 0) or 0)
                    if (min_pos - eps <= item_pos <= max_pos + eps and
                            min_layer <= item_layer <= max_layer):
                        item_type = "transition" if isinstance(item, Transition) else "clip"
                        matched_items.append((item.id, item_type))
                except (TypeError, ValueError):
                    pass

            if not matched_items:
                # Fallback: always keep clicked item selected for this SHIFT action.
                matched_items = [(clicked_item.id, sel_type)]

            if not ctrl:
                self.win.selected_items = []

            for item_id, item_type in matched_items:
                self.win.addSelection(str(item_id), item_type, False)

            self.clip_painter.clear_render_cache()
            self.geometry.mark_dirty()
            self._keyframes_dirty = True
            self.update()

            # Keep the original anchor during SHIFT range operations.
            # This prevents repeated SHIFT press events from collapsing the
            # selection to a single clicked item.
            self.changed(None)
            # Fall through to drag setup; SHIFT+drag freezes horizontal movement.

        elif ctrl and already:
            dbl = QApplication.doubleClickInterval() / 1000.0
            just_added = (
                getattr(self, "_ctrl_just_selected_id", None) == clicked_item.id and
                time.monotonic() - getattr(self, "_ctrl_just_selected_time", 0.0) < dbl
            )
            self._ctrl_just_selected_id = None
            self._last_click_pos = self._selection_anchor_from_item(clicked_item, sel_type)
            if just_added:
                # Second press of a double-click: clip was just CTRL-added;
                # don't immediately toggle it back off — preserve selection.
                pass
            else:
                # Deliberate CTRL+Click on selected item: toggle off.
                self._ctrl_just_deselected_id = clicked_item.id
                self._ctrl_just_deselected_time = time.monotonic()
                self._deselect_timeline_item(clicked_item.id, sel_type)
                self.dragging_items = []
                self._drag_transaction_id = None
                self.changed(None)
                return

        elif not already:
            dbl = QApplication.doubleClickInterval() / 1000.0
            just_deselected = (
                ctrl and
                getattr(self, "_ctrl_just_deselected_id", None) == clicked_item.id and
                time.monotonic() - getattr(self, "_ctrl_just_deselected_time", 0.0) < dbl
            )
            self._ctrl_just_deselected_id = None
            if just_deselected:
                # Second press of a double-click: clip was just CTRL-deselected;
                # don't immediately re-add it — preserve deselected state.
                self._last_click_pos = self._selection_anchor_from_item(clicked_item, sel_type)
            else:
                # Regular click clears+selects; CTRL+click adds to selection.
                self.win.addSelection(clicked_item.id, sel_type, not ctrl)
                if ctrl:
                    self._ctrl_just_selected_id = clicked_item.id
                    self._ctrl_just_selected_time = time.monotonic()
                else:
                    self._ctrl_just_selected_id = None
                self._last_click_pos = self._selection_anchor_from_item(clicked_item, sel_type)
                self.changed(None)

        else:
            # Clicking already-selected item (no special modifier):
            # preserve multi-selection for group drag, but collapse to this
            # one item if this ends as a plain click (no drag movement).
            self._ctrl_just_selected_id = None
            self._last_click_pos = self._selection_anchor_from_item(clicked_item, sel_type)
            selected_count = (
                len(getattr(self.win, "selected_clips", []) or [])
                + len(getattr(self.win, "selected_transitions", []) or [])
            )
            if selected_count > 1:
                self._collapse_selection_on_release = True
                self._collapse_selection_target = (clicked_item.id, sel_type)

        # All selected clips and transitions participate in the drag
        self.dragging_items = [
            itm
            for _rect, itm, selected, _type in self.geometry.iter_items(viewport=False)
            if selected
        ]
        if not self.dragging_items:
            self.dragging_items = [clicked_item]
        if self._selection_overlaps_locked_tracks(self.dragging_items):
            self.dragging_items = []
            self._drag_transaction_id = None
            self._release_cursor()
            return

        # Map track number → index
        self._track_index_from_num = {
            self.normalize_track_number(t.data["number"]): idx
            for idx, t in enumerate(self.track_list)
        }
        self._track_num_from_index = {
            idx: self.normalize_track_number(t.data["number"])
            for idx, t in enumerate(self.track_list)
        }

        # Record each item’s starting position and layer index
        fps = float(self.fps_float or 0.0)
        use_frames = fps > 0.0
        self._drag_initial = {}
        for itm in self.dragging_items:
            data = itm.data if isinstance(itm.data, dict) else {}
            position = float(data.get("position", 0.0) or 0.0)
            start = float(data.get("start", 0.0) or 0.0)
            end = float(data.get("end", start) or start)
            duration = max(0.0, end - start)
            index = self._track_index_from_num.get(data.get("layer", 0), 0)

            entry = {
                "position": position,
                "index": index,
                "duration": duration,
            }

            if use_frames:
                entry["position_frames"] = int(round(position * fps))
                entry["duration_frames"] = int(round(duration * fps))

            self._drag_initial[itm.id] = entry

        # Seed pending overrides so geometry rebuilds use drag positions
        for itm in self.dragging_items:
            if isinstance(itm, Clip):
                override = self._pending_clip_overrides.setdefault(itm.id, {})
                override["position"] = float(itm.data.get("position", 0.0) or 0.0)
                override.setdefault("start", float(itm.data.get("start", 0.0) or 0.0))
                override.setdefault("end", float(itm.data.get("end", 0.0) or 0.0))
                override["layer"] = itm.data.get("layer", 0)
            elif isinstance(itm, Transition):
                override = self._pending_transition_overrides.setdefault(itm.id, {})
                override["position"] = float(itm.data.get("position", 0.0) or 0.0)
                override["start"] = float(itm.data.get("start", 0.0) or 0.0)
                override["end"] = float(itm.data.get("end", 0.0) or 0.0)
                override["layer"] = itm.data.get("layer", 0)

        # Bounding box for snapping calculations
        self.drag_bbox = self._compute_selected_bounding()

        # Horizontal offset from cursor to bbox-left
        self.drag_clip_offset = e.pos().x() - self.drag_bbox.x()

        # Starting track index
        start_idx = self._track_index_at_viewport_y(
            e.pos().y(),
            prefer_clip_lane=True,
            snap_to_nearest=True,
        )
        self._drag_layer_idx_start = start_idx if start_idx is not None else 0

    def _dragMove(self):
        """Apply identical horizontal/vertical deltas to every dragged item."""
        if getattr(self, "_drag_commit_in_progress", False):
            return
        if not getattr(self, "dragging_items", None):
            return
        e = self._last_event

        if not getattr(self, "_drag_threshold_met", True):
            anchor = getattr(self, "_drag_press_pos", None)
            if anchor is not None and e is not None:
                delta = e.pos() - anchor
                if delta.manhattanLength() < QApplication.startDragDistance():
                    return
            self._drag_threshold_met = True

        # -------- Horizontal delta (seconds) --------
        pps = float(self.pixels_per_second or 0.0)
        if pps <= 0.0:
            return

        # SHIFT+Drag: freeze horizontal movement (track-only drag), matching JS behaviour.
        shift_held = bool(e.modifiers() & Qt.ShiftModifier) if e else False
        if shift_held:
            delta_sec = 0.0
        else:
            new_bbox_x = e.pos().x() - self.drag_clip_offset
            delta_sec = (new_bbox_x - self.drag_bbox.x()) / pps

            # Snap horizontally ±1.5 s (pure x-axis)
            if self.enable_snapping:
                delta_sec = self._snap_delta(delta_sec)

        # -------- Vertical delta (track indexes) ----
        new_idx_under_cursor = self._track_index_at_viewport_y(
            e.pos().y(),
            prefer_clip_lane=True,
            snap_to_nearest=True,
        )
        if new_idx_under_cursor is None:
            new_idx_under_cursor = self._drag_layer_idx_start
        delta_idx = new_idx_under_cursor - self._drag_layer_idx_start

        # Clamp delta_idx so *all* items stay within valid index range
        orig_indices = [info["index"] for info in self._drag_initial.values()]
        if orig_indices:
            if min(orig_indices) + delta_idx < 0:
                delta_idx = -min(orig_indices)
            if max(orig_indices) + delta_idx >= len(self.track_list):
                delta_idx = (len(self.track_list) - 1) - max(orig_indices)

        # Clamp horizontal delta so items do not move before t=0.
        start_positions = [info["position"] for info in self._drag_initial.values()]
        if start_positions:
            min_delta_sec = -min(start_positions)
            if delta_sec < min_delta_sec:
                delta_sec = min_delta_sec

        fps = float(self.fps_float or 0.0)
        frame_offset = None
        if fps > 0.0:
            frame_offset = int(round(delta_sec * fps))

            start_frames = [
                info.get("position_frames")
                for info in self._drag_initial.values()
                if info.get("position_frames") is not None
            ]
            if start_frames:
                min_frame_offset = -min(start_frames)
                if frame_offset < min_frame_offset:
                    frame_offset = min_frame_offset

            delta_sec = frame_offset / fps

        # Reapply left bound to account for frame rounding
        if start_positions:
            min_delta_sec = -min(start_positions)
            if delta_sec < min_delta_sec:
                delta_sec = min_delta_sec

        # -------- Apply identical deltas ------------
        for itm in self.dragging_items:
            info = self._drag_initial[itm.id]
            start_pos_sec = info["position"]
            start_idx = info["index"]

            # New values
            if frame_offset is not None:
                new_pos_sec = start_pos_sec + (frame_offset / fps)
            else:
                new_pos_sec = start_pos_sec + delta_sec
            new_pos_sec = max(0.0, new_pos_sec)
            new_idx = start_idx + delta_idx
            new_idx = max(0, min(new_idx, len(self.track_list) - 1))
            unlocked_idx = self._nearest_unlocked_track_index(new_idx)
            if unlocked_idx is None:
                unlocked_idx = start_idx
            new_idx = unlocked_idx
            new_layer_num = self._track_num_from_index[new_idx]

            if (
                not getattr(self, "_drag_moved", False)
                and (
                    abs(new_pos_sec - start_pos_sec) > 1e-6
                    or new_idx != start_idx
                )
            ):
                self._drag_moved = True

            itm.data["position"] = new_pos_sec
            itm.data["layer"] = new_layer_num

            if isinstance(itm, Clip):
                override = self._pending_clip_overrides.setdefault(itm.id, {})
            else:
                override = self._pending_transition_overrides.setdefault(itm.id, {})
            override["position"] = new_pos_sec
            override["layer"] = new_layer_num

            # Update cached rect
            rect = self.geometry.calc_item_rect(itm)
            self.geometry.update_item_rect(itm, rect)
            # Use the actual applied movement (after per-item clamping/snap-to-frame),
            # otherwise panel points can lag at the snap threshold while clip visuals
            # already snapped to the final frame-aligned position.
            applied_delta_sec = new_pos_sec - start_pos_sec
            if fps > 0.0:
                applied_frame_delta = int(round(applied_delta_sec * fps))
            else:
                applied_frame_delta = 0
            # Always apply panel shift, even for 0 delta. When snapping returns to
            # drag origin, skipping this leaves stale panel points until mouse-up.
            self._panel_shift_item(itm, applied_delta_sec, applied_frame_delta)

        # Immediate visual feedback
        self._keyframes_dirty = True
        self.update()

    def _finishClipDrag(self):
        """Persist all moved clips/transitions and refresh geometry."""
        items = list(getattr(self, "dragging_items", None) or [])
        moved = bool(getattr(self, "_drag_moved", False))
        collapse_selection = bool(getattr(self, "_collapse_selection_on_release", False))
        collapse_target = getattr(self, "_collapse_selection_target", None)

        if items and moved:
            # Freeze the drag state before dispatching any updates. Otherwise a
            # late mouse-move can still mutate the remaining selected items
            # while this save loop is running, causing them to commit to
            # different layers/positions within one drag.
            transaction_id = self._drag_transaction_id
            self._drag_commit_in_progress = True
            self.dragging_items = []
            self._drag_transaction_id = None
            self._preserve_overrides_once = True
            total = len(items)
            commit_items = []
            for itm in items:
                commit_items.append({
                    "item": itm,
                    "type": "transition" if isinstance(itm, Transition) else "clip",
                    "data": json.loads(json.dumps(itm.data)),
                })
            # Persist clips before transitions so any transition logic that
            # inspects neighboring clips sees the final clip positions.
            commit_items.sort(key=lambda entry: 1 if entry["type"] == "transition" else 0)
            try:
                for idx, entry in enumerate(commit_items):
                    itm = entry["item"]
                    ignore_refresh = idx < total - 1
                    if entry["type"] == "transition":
                        transition_data = entry["data"]
                        if total == 1 and self._transition_uses_static_mask(transition_data):
                            transition_data["_auto_direction"] = True
                        self.update_transition_data(
                            transition_data,
                            only_basic_props=True,
                            ignore_refresh=ignore_refresh,
                            transaction_id=transaction_id,
                        )
                        itm.data = transition_data
                    else:
                        clip_data = entry["data"]
                        if total == 1:
                            clip_data["_auto_transition"] = True
                        self.update_clip_data(
                            clip_data,
                            only_basic_props=True,
                            ignore_reader=True,
                            ignore_refresh=ignore_refresh,
                            transaction_id=transaction_id,
                        )
                        itm.data = clip_data
            finally:
                self._drag_commit_in_progress = False
        elif items and not moved:
            for itm in items:
                if isinstance(itm, Transition):
                    self._pending_transition_overrides.pop(itm.id, None)
                else:
                    self._pending_clip_overrides.pop(itm.id, None)

        self.dragging_items = []
        self._drag_transaction_id = None
        self.snap.reset()
        self._collapse_selection_on_release = False
        self._collapse_selection_target = None
        if moved:
            self._update_project_duration()
            self.changed(None)
        else:
            if collapse_selection and collapse_target:
                target_id, target_type = collapse_target
                self.win.addSelection(target_id, target_type, True)
                self.changed(None)
            self.geometry.mark_dirty()
        self.update()
        self._release_cursor()
        if self._last_event:
            self._updateCursor(self._last_event.pos())
        self._drag_moved = False
        self._drag_press_pos = None
        self._drag_threshold_met = False

    def _compute_selected_bounding(self):
        """Return a QRectF encompassing all currently-selected clips and transitions."""
        rects = [
            rect
            for rect, _item, selected, _type in self.geometry.iter_items(viewport=False)
            if selected
        ]
        if not rects:
            return QRectF()
        bbox = QRectF(rects[0])
        for rect in rects[1:]:
            bbox = bbox.united(rect)
        return bbox

    def _snap_delta(self, delta_seconds):
        """
        Given a proposed horizontal delta (seconds) for the group drag, adjust it
        so the selection’s left or right edge “snaps” to the nearest clip edge
        within ±1.5 seconds.  Snapping is strictly horizontal—layer movement is
        unaffected.
        """
        original_ignore = getattr(self, "_snap_ignore_ids", set())
        try:
            ignore_ids = {
                getattr(item, "id", None)
                for item in getattr(self, "dragging_items", [])
            }
            self._snap_ignore_ids = {obj_id for obj_id in ignore_ids if obj_id is not None}
            return self.snap.snap_dx(delta_seconds)
        finally:
            self._snap_ignore_ids = original_ignore

    def _snap_trim_delta(self, delta_seconds, edge=None, initial=None):
        """
        Apply directional edge snapping to trim deltas.

        Uses the trim edge's original timeline position (left/right) and lets
        SnapHelper.snap_edge() handle direction-aware target selection.
        """
        if initial is None:
            initial = getattr(self, "_resize_initial", None)
        if not isinstance(initial, dict):
            return delta_seconds

        edge_label = edge or getattr(self, "_resize_edge", None)
        if edge_label not in ("left", "right"):
            return delta_seconds

        if edge_label == "left":
            edge_sec = float(initial.get("position", 0.0) or 0.0)
        else:
            edge_sec = float(initial.get("position", 0.0) or 0.0)
            edge_sec += float(initial.get("end", 0.0) or 0.0) - float(initial.get("start", 0.0) or 0.0)

        return self.snap.snap_edge(edge_sec, delta_seconds)

    def _startResize(self):
        if self._press_hit == "clip-edge" and self._resizing_item:
            self._startItemResize()
        elif self._press_hit == "timeline-handle":
            self._startProjectResize()
        elif self._press_hit == "handle":
            self._resize_start = self.track_name_width

    def _resizeMove(self):
        if self._press_hit == "clip-edge" and self._resizing_item:
            self._itemResizeMove()
        elif self._press_hit == "timeline-handle":
            self._projectResizeMove()
        elif self._press_hit == "handle":
            new_width = max(40, self._last_event.pos().x())
            if new_width != self.track_name_width:
                self.track_name_width = new_width
                self.changed(None)

    def _finishResize(self):
        if self._press_hit == "clip-edge" and self._resizing_item:
            self._finishItemResize()
            if hasattr(self.win, "TimelinePreviewMode"):
                self.win.TimelinePreviewMode.emit()
        elif self._press_hit == "timeline-handle":
            self._finishProjectResize()
        else:
            pass

    def _startProjectResize(self):
        self._fix_cursor(self.cursors.get("resize_x", Qt.SizeHorCursor))
        self._project_resize_initial_duration = self._current_project_duration()
        min_duration = self._furthest_timeline_edge()
        if self.fps_float > 0:
            min_duration = max(min_duration, 1.0 / self.fps_float)
        self._project_resize_min_duration = max(0.0, min_duration)
        self._project_resize_keep_right = self._is_view_right_aligned()
        self._set_project_duration_override(self._project_resize_initial_duration)

    def _projectResizeMove(self):
        event = self._last_event
        if not event:
            return
        new_duration = self._seconds_from_x(event.pos().x())
        new_duration = max(self._project_resize_min_duration, new_duration)
        snapped = self._snap_time(new_duration)
        if snapped < self._project_resize_min_duration:
            snapped = self._project_resize_min_duration
        current_preview = getattr(self, "_project_duration_override", None)
        if current_preview is None or abs(snapped - current_preview) > 1e-4:
            self._set_project_duration_override(snapped)

    def _finishProjectResize(self):
        final_duration = getattr(self, "_project_duration_override", None)
        if final_duration is None:
            final_duration = self._project_resize_initial_duration
        final_duration = max(self._project_resize_min_duration, float(final_duration or 0.0))
        snapped = self._snap_time(final_duration)
        if snapped < self._project_resize_min_duration:
            snapped = self._project_resize_min_duration
        self._set_project_duration_override(None)
        self._project_resize_keep_right = False
        self._release_cursor()
        if abs(snapped - self._project_resize_initial_duration) <= 1e-3:
            return
        timeline = getattr(self.win, "timeline", None)
        if timeline:
            timeline.resizeTimeline(snapped)

    def _startItemResize(self):
        item = self._resizing_item
        if not item:
            return
        if self._is_track_locked((item.data if isinstance(item.data, dict) else {}).get("layer")):
            self._resizing_item = None
            self._resize_items = []
            self._resize_initial_map = {}
            self._resize_edge = None
            self._release_cursor()
            return
        if hasattr(self.win, "TrimPreviewMode"):
            self.win.TrimPreviewMode.emit()
        self.snap.reset()
        self._fix_cursor(self.cursors["resize_x"])
        resize_items = list(getattr(self, "_resize_items", None) or [item])
        self._resize_items = resize_items
        self._resize_initial_map = {}
        primary_context = None
        self._resize_clip_max_duration = None
        self._resize_clip_is_single_image = False
        self._resize_allow_left_overflow = False
        self._resize_snap_ignore_backup = set(getattr(self, "_snap_ignore_ids", set()))
        updated_ignore = set(self._resize_snap_ignore_backup)
        for candidate in resize_items:
            candidate_id = getattr(candidate, "id", None)
            if candidate_id is not None:
                updated_ignore.add(candidate_id)
            context = self._build_resize_context(candidate)
            self._resize_initial_map[candidate.id] = context
            self._apply_resize_preview_override(
                candidate,
                context,
                context["initial"]["start"],
                context["initial"]["end"],
                context["initial"]["position"],
            )
            if isinstance(candidate, Clip):
                self._set_trim_thumbnail_suspension(True, candidate.id)
            if candidate is item:
                primary_context = context
        self._snap_ignore_ids = updated_ignore
        if primary_context is None:
            primary_context = self._build_resize_context(item)
            self._resize_initial_map[item.id] = primary_context

        self._resize_initial_world_rect = QRectF(primary_context["world_rect"])
        self._resize_initial_rect = QRectF(primary_context["rect"])
        self._resize_initial = dict(primary_context["initial"])
        self._resize_clip_max_duration = primary_context.get("max_duration")
        self._resize_clip_is_single_image = bool(primary_context.get("clip_is_single_image"))
        self._resize_allow_left_overflow = bool(primary_context.get("allow_left_overflow"))
        if isinstance(item, Clip):
            self._timing_original_start = self._resize_initial["start"]
            sel_type = "clip"
        else:
            sel_type = "transition"
            self._snap_keyframe_seconds = []
        # Ensure item is selected
        self.win.addSelection(item.id, sel_type, False)

        if isinstance(item, Clip) and not self.enable_timing:
            # Rebuild markers before collecting trim snap targets.
            self._keyframes_dirty = True
            self._update_snap_keyframe_targets(item)

    def _itemResizeMove(self):
        resize_items = list(getattr(self, "_resize_items", None) or [])
        if not resize_items:
            return
        self._resize_results = {}
        preview_item = self._resize_preview_focus_item()
        preview_result = None
        primary_item = self._resizing_item if self._resizing_item in resize_items else resize_items[0]
        primary_context = self._resize_initial_map.get(getattr(primary_item, "id", None))
        shared_edge_seconds = None
        primary_result = None
        if primary_context:
            if isinstance(primary_item, Transition):
                primary_result = self._compute_transition_resize(primary_item, primary_context)
            else:
                primary_result = self._compute_clip_resize(primary_item, primary_context)
            _rect, start, end, position = primary_result
            shared_edge_seconds = self._resize_result_edge_seconds(start, end, position)

        for item in resize_items:
            context = self._resize_initial_map.get(item.id)
            if not context:
                continue
            if item is primary_item and primary_result is not None:
                rect, start, end, position = primary_result
            elif isinstance(item, Transition):
                rect, start, end, position = self._compute_transition_resize(
                    item,
                    context,
                    target_edge_seconds=shared_edge_seconds,
                )
            else:
                rect, start, end, position = self._compute_clip_resize(
                    item,
                    context,
                    target_edge_seconds=shared_edge_seconds,
                )

            self._resize_results[item.id] = {
                "start": start,
                "end": end,
                "position": position,
            }
            self.geometry.update_item_rect(item, rect)
            self._apply_resize_preview_override(item, context, start, end, position)
            self._keyframes_dirty = True

            if item is self._resizing_item:
                self._resize_new_start = start
                self._resize_new_end = end
                self._resize_new_position = position
            if item is preview_item:
                preview_result = {
                    "item": item,
                    "start": start,
                    "end": end,
                    "position": position,
                }
            if isinstance(item, Clip) and self.enable_timing:
                self._snap_keyframe_seconds = []

        timeline = getattr(self.win, "timeline", None)
        if preview_result and timeline and self.fps_float:
            item = preview_result["item"]
            start = preview_result["start"]
            end = preview_result["end"]
            if isinstance(item, Clip):
                item_id = getattr(item, "id", None)
                if item_id:
                    if self._resize_edge == "left":
                        frame_seconds = self._snap_time(start)
                    else:
                        frame_seconds = self._snap_time(end)
                    frame = int(round(frame_seconds * self.fps_float)) + 1
                    timeline.PreviewClipFrame(str(item_id), max(1, frame))
            elif isinstance(item, Transition):
                item_id = getattr(item, "id", None)
                if item_id:
                    if self._resize_edge == "left":
                        frame_seconds = self._snap_time(start)
                    else:
                        frame_seconds = self._snap_time(end)
                    frame = int(round(frame_seconds * self.fps_float)) + 1
                    timeline.PreviewTransitionFrame(str(item_id), max(1, frame))
        self.update()

    def _resize_result_edge_seconds(self, start, end, position):
        if getattr(self, "_resize_edge", None) == "left":
            return position
        return position + max(0.0, end - start)

    def _compute_transition_resize(self, item, context=None, target_edge_seconds=None):
        event = self._last_event
        pps = self.pixels_per_second
        min_len = 1.0 / self.fps_float
        if context is None:
            context = {
                "rect": self._resize_initial_rect,
                "world_rect": getattr(self, "_resize_initial_world_rect", self._resize_initial_rect),
                "initial": self._resize_initial,
                "static_mask": self._transition_uses_static_mask(item.data),
            }
        rect = context["rect"]
        world_rect = context.get("world_rect", rect)
        initial = context["initial"]
        start = initial["start"]
        end = initial["end"]
        width = max(end - start, min_len)
        pos = initial["position"]
        static_mask = bool(context.get("static_mask"))

        if self._resize_edge == "left":
            if target_edge_seconds is not None:
                delta_sec = target_edge_seconds - pos
            else:
                delta_sec = (event.pos().x() - rect.left()) / pps
            if target_edge_seconds is None and self.enable_snapping:
                delta_sec = self.snap.snap_edge(pos, delta_sec)
            max_delta = width - min_len
            if delta_sec > max_delta:
                delta_sec = max_delta
            new_position = pos + delta_sec
            new_start = 0.0 if static_mask else start + delta_sec
            new_end = (width - delta_sec) if static_mask else end
            if new_position < 0:
                new_position = 0
                if static_mask:
                    new_end = (pos + width) - new_position
                else:
                    new_start = start + (new_position - pos)
            rect_left = self.track_name_width + new_position * pps
        else:
            if target_edge_seconds is not None:
                delta_sec = target_edge_seconds - (pos + width)
            else:
                delta_sec = (event.pos().x() - rect.right()) / pps
            if target_edge_seconds is None and self.enable_snapping:
                delta_sec = self.snap.snap_edge(pos + width, delta_sec)
            min_delta = -(width - min_len)
            if delta_sec < min_delta:
                delta_sec = min_delta
            new_start = 0.0 if static_mask else start
            new_end = (width + delta_sec) if static_mask else end + delta_sec
            new_position = pos
            rect_left = self.track_name_width + new_position * pps

        rect_width = max(new_end - new_start, min_len) * pps
        geom_rect = QRectF(rect_left, world_rect.y(), rect_width, world_rect.height())
        return geom_rect, new_start, new_end, new_position

    def _compute_clip_resize(self, item, context=None, target_edge_seconds=None):
        event = self._last_event
        pps = float(self.pixels_per_second or 0.0)
        if context is None:
            context = {
                "rect": self._resize_initial_rect,
                "world_rect": getattr(self, "_resize_initial_world_rect", self._resize_initial_rect),
                "initial": self._resize_initial,
                "max_duration": getattr(self, "_resize_clip_max_duration", None),
                "allow_left_overflow": bool(getattr(self, "_resize_allow_left_overflow", False)),
                "clip_is_single_image": bool(getattr(self, "_resize_clip_is_single_image", False)),
            }
        rect = context["rect"]
        world_rect = context.get("world_rect", rect)
        initial = context["initial"]
        start = initial["start"]
        end = initial["end"]
        pos = initial["position"]
        duration = initial["duration"]
        fps = self.fps_float or 1.0
        min_len = 1.0 / fps
        max_duration = context.get("max_duration")
        allow_left_overflow = bool(context.get("allow_left_overflow", False))
        single_image_resize = bool(context.get("clip_is_single_image", False))
        overflow_enabled = allow_left_overflow and single_image_resize and not self.enable_timing

        if target_edge_seconds is None and (event is None or pps <= 0.0):
            geom_rect = QRectF(world_rect)
            return geom_rect, start, end, pos

        cursor_sec = target_edge_seconds
        if cursor_sec is None:
            cursor_sec = self._seconds_from_x(event.pos().x())
        clip_span = max(end - start, min_len)

        if self._resize_edge == "left":
            delta_sec = cursor_sec - pos
            if target_edge_seconds is None and self.enable_snapping:
                delta_sec = self._snap_trim_delta(delta_sec, edge="left", initial=initial)
            if overflow_enabled and max_duration is not None:
                extra_capacity = max(0.0, max_duration - end)
                min_delta = -start - extra_capacity
                if delta_sec < min_delta:
                    delta_sec = min_delta
            new_position = pos + delta_sec
            new_start = start + delta_sec
            new_end = end

            max_start = end - min_len
            overflow = 0.0
            if new_start < 0.0 and not self.enable_timing:
                overflow = -new_start
                new_start = 0.0
                if not allow_left_overflow and not self.enable_timing:
                    new_position = pos - start
            if new_start > max_start:
                new_start = max_start
                new_position = pos + (max_start - start)
            if new_position < 0.0 and not self.enable_timing:
                diff = -new_position
                new_position = 0.0
                new_start += diff
            if overflow > 0.0 and overflow_enabled:
                target_end = new_end + overflow
                if max_duration is not None and target_end > max_duration:
                    target_end = max_duration
                new_end = target_end
            rect_left = self.track_name_width + new_position * pps
        else:
            timeline_right = pos + clip_span
            delta_sec = cursor_sec - timeline_right
            if target_edge_seconds is None and self.enable_snapping:
                delta_sec = self._snap_trim_delta(delta_sec, edge="right", initial=initial)
            new_end = end + delta_sec
            new_start = start
            new_position = pos

            min_end = start + min_len
            if new_end < min_end:
                new_end = min_end
            if not self.enable_timing:
                max_end = max_duration
                if max_end is None:
                    max_end = start + duration
                if new_end > max_end:
                    new_end = max_end
            rect_left = self.track_name_width + new_position * pps

        rect_width = (new_end - new_start) * pps
        geom_rect = QRectF(rect_left, world_rect.y(), rect_width, world_rect.height())
        return geom_rect, new_start, new_end, new_position

    def _finishItemResize(self):
        item = self._resizing_item
        resize_items = list(getattr(self, "_resize_items", None) or ([item] if item else []))
        if not item:
            return
        if not hasattr(self, "_resize_new_start"):
            for candidate in resize_items:
                if isinstance(candidate, Clip):
                    self._set_trim_thumbnail_suspension(False, candidate.id)
                else:
                    # Resize can start/end without a move event; clear any preview
                    # override seeded in _startItemResize().
                    self._pending_transition_overrides.pop(candidate.id, None)
                self._clear_resize_preview_overrides(candidate)
            self._resizing_item = None
            self._resize_items = []
            self._resize_initial_map = {}
            self._snap_keyframe_seconds = []
            self.snap.reset()
            self._restore_resize_snap_ignore_ids(resize_items)
            # Ensure selection visuals are fully refreshed even when resize
            # starts/ends without movement (click on edge).
            self.changed(None)
            self.geometry.mark_dirty()
            self._keyframes_dirty = True
            self.update()
            self._release_cursor()
            if self._last_event:
                self._updateCursor(self._last_event.pos())
            return
        resize_results = getattr(self, "_resize_results", {}) or {}
        resize_initial_map = dict(getattr(self, "_resize_initial_map", {}) or {})
        resize_edge = self._resize_edge
        self._resizing_item = None
        self._resize_items = []
        self._resize_initial_map = {}
        self._resize_results = {}
        self._keyframes_dirty = True
        suspend_keyframe_rebuild = (
            len(resize_items) > 1
            and any(isinstance(candidate, Transition) for candidate in resize_items)
        )
        if suspend_keyframe_rebuild:
            self._suspend_keyframe_rebuild = True
        transaction_id = str(uuid.uuid4())
        total = len(resize_items)
        waveform_clips = []
        requested_backend_refresh = False
        self._suspend_changed_update += 1
        self._preserve_overrides_during_batch = True
        try:
            for idx, candidate in enumerate(resize_items):
                result = resize_results.get(candidate.id)
                if not isinstance(result, dict):
                    continue
                start = result["start"]
                end = result["end"]
                position = result["position"]
                setattr(self.win, "_trim_refresh_pending", True)
                ignore_refresh = idx < total - 1
                if not ignore_refresh:
                    requested_backend_refresh = True
                if isinstance(candidate, Clip):
                    context = resize_initial_map.get(candidate.id, {})
                    audio_data = candidate.data.get("ui", {}).get("audio_data")
                    has_waveform = audio_data not in (None, [])
                    if self._commit_resized_clip(
                        candidate,
                        start,
                        end,
                        position,
                        context,
                        transaction_id,
                        ignore_refresh,
                    ) and has_waveform:
                        waveform_clips.append(candidate.id)
                else:
                    context = resize_initial_map.get(candidate.id, {})
                    static_mask = bool(context.get("static_mask"))
                    transition_data = json.loads(json.dumps(candidate.data))
                    start, end, position = self._normalize_resize_commit_bounds(start, end, position)
                    transition_data["position"] = position
                    transition_data["start"] = start
                    transition_data["end"] = end
                    transition_data["duration"] = self._snap_time(transition_data["end"] - transition_data["start"])
                    transition_data["_auto_direction"] = static_mask
                    self.update_transition_data(
                        transition_data,
                        only_basic_props=True,
                        ignore_refresh=ignore_refresh,
                        transaction_id=transaction_id,
                    )
                    candidate.data = transition_data

                if isinstance(candidate, (Clip, Transition)) and hasattr(self, "RefreshTrimmedTimelineItem"):
                    self.RefreshTrimmedTimelineItem(json.dumps(candidate.data), resize_edge)

                if isinstance(candidate, Clip):
                    self._set_trim_thumbnail_suspension(False, candidate.id)
        finally:
            self._preserve_overrides_during_batch = False
            self._suspend_changed_update = max(0, self._suspend_changed_update - 1)
            self._suspend_keyframe_rebuild = False

        self._snap_keyframe_seconds = []
        self.snap.reset()
        self._restore_resize_snap_ignore_ids(resize_items)
        self._update_project_duration()
        if waveform_clips and hasattr(self, "Show_Waveform_Triggered"):
            self.Show_Waveform_Triggered(waveform_clips, transaction_id=transaction_id)
        self._finalize_resize_preview_state(resize_items, requested_backend_refresh)
        if hasattr(self, "_refresh_keyframe_markers") and not self._dragging_panel_keyframes and not self._dragging_keyframe:
            self._refresh_keyframe_markers()
        else:
            self._keyframes_dirty = True
        self.update()
        self._release_cursor()
        if self._last_event:
            self._updateCursor(self._last_event.pos())
        if hasattr(self, "_resize_initial_world_rect"):
            del self._resize_initial_world_rect
        self._resize_clip_max_duration = None
        self._resize_allow_left_overflow = False
        self._resize_clip_is_single_image = False

    def _startBoxSelect(self):
        e = self._last_event
        ctrl_down = bool(e.modifiers() & Qt.ControlModifier)
        self.box_start = e.pos()
        panel_lane = self._panel_lane_at(self.box_start)
        panel_track = panel_lane.get("track") if panel_lane else self._panel_track_at_pos(self.box_start)
        if panel_track is not None:
            self._panel_box_track = panel_track
            self._panel_box_bounds = self._panel_bounds_for_track(self._panel_box_track)
            if not ctrl_down:
                self._clear_panel_selection(self._panel_box_track)
        else:
            self._panel_box_track = None
            self._panel_box_bounds = QRectF()
            if not ctrl_down:
                # Starting a new box selection clears existing selections
                self.win.clearSelections()
        self.selection_rect = QRectF()

    def _boxMove(self):
        rect = QRectF(self.box_start, self._last_event.pos()).normalized()
        if self._panel_box_track is not None:
            bounds = self._panel_box_bounds
            if isinstance(bounds, QRectF) and not bounds.isNull():
                rect = rect.intersected(bounds)
            else:
                rect = QRectF()
        self.selection_rect = rect
        self.update()

    def _finishBoxSelect(self):
        """Finalize box-select: add items intersecting the selection rectangle."""
        self.geometry.ensure()
        if self._panel_box_track is not None:
            ctrl_down = False
            if self._last_event and hasattr(self._last_event, "modifiers"):
                mods = self._last_event.modifiers()
                ctrl_down = bool(mods & Qt.ControlModifier)
            track_num = self._panel_box_track
            selection_rect = QRectF(self.selection_rect)
            frames_by_prop = {}
            if not selection_rect.isNull():
                for lane in self._iter_panel_lanes() or []:
                    if lane.get("track") != track_num:
                        continue
                    combined = lane.get("combined_rect", QRectF())
                    if combined.isNull() or not combined.intersects(selection_rect):
                        continue
                    prop = lane.get("property") or {}
                    points = prop.get("points") or []
                    if not points:
                        continue
                    lane_rect = lane.get("lane_rect", QRectF())
                    lane_padding = lane.get("lane_padding", self._panel_lane_padding())
                    selected_frames = set()
                    selected_context = None
                    for point in points:
                        seconds = point.get("seconds")
                        if seconds is None:
                            continue
                        marker_rect = self._panel_marker_rect(lane_rect, lane_padding, seconds)
                        if marker_rect.intersects(selection_rect):
                            point_context = None
                            if hasattr(self, "_panel_property_context"):
                                point_context = self._panel_property_context(
                                    {"context": point.get("_panel_context")},
                                    lane.get("context"),
                                )
                            if selected_context is None and point_context:
                                selected_context = self._panel_context_signature(point_context)
                            frame_val = point.get("frame")
                            if frame_val is not None:
                                try:
                                    selected_frames.add(int(frame_val))
                                except (TypeError, ValueError):
                                    continue
                    if selected_frames:
                        selector = self._panel_selection_selector(
                            selected_frames,
                            context_signature=selected_context,
                        )
                        frames_by_prop[prop.get("key")] = selector
            if ctrl_down:
                if frames_by_prop:
                    self._panel_merge_selection_map(track_num, frames_by_prop)
            else:
                self._panel_set_selection_map(track_num, frames_by_prop)
            self.selection_rect = QRectF()
            self._panel_box_track = None
            self._panel_box_bounds = QRectF()
            self.update()
            return

        # Ensure geometry is up-to-date for clip selections
        self.geometry.mark_dirty()
        self.geometry.ensure()

        # Add any item whose rect intersects selection_rect
        for rect, item, _selected, _type in self.geometry.iter_items():
            if rect.intersects(self.selection_rect):
                sel_type = "transition" if isinstance(item, Transition) else "clip"
                # False = don’t emit SelectionChanged (we’ll handle it ourselves)
                self.win.addSelection(item.id, sel_type, False)

        # Clear the box
        self.selection_rect = QRectF()

        # Recompute all clip/track geometry and repaint immediately
        self.changed(None)
        self.update()



"""
 @file
 @brief This file contains unit tests for project data loading and migration
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import sys
import tempfile
import types
import unittest
from contextlib import ExitStack
from unittest.mock import patch


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from PyQt5.QtWidgets import QApplication

from classes.project_data import ProjectDataStore
from classes.updates import UpdateManager
from qt_test_app import ensure_app_state as ensure_qt_app_state, get_or_create_app


class DummySettings:
    actionType = types.SimpleNamespace(IMPORT="import")

    def __init__(self):
        self.values = {
            "recent_projects": [],
            "default-profile": "HD 720p 30 fps",
            "default-samplerate": 48000,
            "default-channels": 2,
        }
        self.saved = False
        self.default_paths = {}

    def get(self, key):
        return self.values.get(key)

    def set(self, key, value):
        self.values[key] = value

    def save(self):
        self.saved = True

    def setDefaultPath(self, action, value):
        self.default_paths[action] = value


class DummyApp(QApplication):
    def __init__(self):
        super().__init__([])
        self.settings = DummySettings()
        self.project = None
        self.updates = None
        self.window = None

    def get_settings(self):
        return self.settings

    def _tr(self, text):
        return text


def ensure_app_state(app):
    return ensure_qt_app_state(
        app,
        DummySettings,
        project_factory=ProjectDataStore,
        updates_factory=UpdateManager,
        extra_attrs={"window": None},
    )


class DummyAction:
    def __init__(self):
        self.enabled = None

    def setEnabled(self, value):
        self.enabled = value


def make_store():
    store = ProjectDataStore.__new__(ProjectDataStore)
    store.data_type = "project data"
    store.current_filepath = None
    store.has_unsaved_changes = False
    store._data = {}
    return store


class ProjectDataTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        app, cls._owns_app = get_or_create_app(DummyApp)
        cls.app = ensure_app_state(app)

    @classmethod
    def tearDownClass(cls):
        if getattr(cls, "_owns_app", False) and cls.app:
            cls.app.quit()

    def setUp(self):
        ensure_app_state(self.app)
        self.app.settings = DummySettings()
        self.app.window = None

    def tearDown(self):
        ensure_app_state(self.app)

    def test_load_restores_history_and_enables_waveform_clear(self):
        store = make_store()
        default_project = {
            "clips": [],
            "effects": [],
            "markers": [],
            "layers": [],
            "files": [],
            "history": {"undo": [], "redo": []},
            "version": {"openshot-qt": "3.4.0", "libopenshot": "0.5.0"},
        }
        loaded_project = {
            "clips": [],
            "effects": [],
            "markers": [],
            "layers": [],
            "files": [{"id": "F1", "ui": {"audio_data": [0.1, 0.2]}}],
            "version": {"openshot-qt": "3.4.0", "libopenshot": "0.5.0"},
        }

        loaded_payloads = []
        clear_waveform_action = DummyAction()
        self.app.window = types.SimpleNamespace(actionClearWaveformData=clear_waveform_action)
        self.app.updates = types.SimpleNamespace(load=loaded_payloads.append)

        with tempfile.TemporaryDirectory() as tmpdir:
            project_path = os.path.join(tmpdir, "example.osp")
            with ExitStack() as stack:
                stack.enter_context(
                    patch.object(store, "new", lambda: setattr(store, "_data", default_project.copy()))
                )
                stack.enter_context(
                    patch.object(
                        store,
                        "read_from_file",
                        lambda file_path, path_mode="ignore": loaded_project.copy(),
                    )
                )
                stack.enter_context(patch.object(store, "check_if_paths_are_valid", lambda: None))
                stack.enter_context(patch.object(store, "add_to_recent_files", lambda file_path: None))
                stack.enter_context(patch.object(store, "upgrade_project_data_structures", lambda: None))
                stack.enter_context(patch.object(store, "get_profile", lambda **kwargs: object()))
                stack.enter_context(patch.object(store, "apply_default_audio_settings", lambda: None))
                ProjectDataStore.load(store, project_path, clear_thumbnails=False)

        self.assertEqual(store.current_filepath, project_path)
        self.assertFalse(store.has_unsaved_changes)
        self.assertEqual(store._data["history"], {"undo": [], "redo": []})
        self.assertTrue(clear_waveform_action.enabled)
        self.assertEqual(loaded_payloads, [store._data])

    def test_load_migrates_flat_thumbnails_into_per_file_folders(self):
        store = make_store()
        project_data = {
            "clips": [],
            "effects": [],
            "markers": [],
            "layers": [],
            "files": [{"id": "F1", "path": "/project/source.mp4"}],
            "history": {"undo": [], "redo": []},
            "version": {"openshot-qt": "3.4.0", "libopenshot": "0.5.0"},
        }

        clear_waveform_action = DummyAction()
        self.app.window = types.SimpleNamespace(actionClearWaveformData=clear_waveform_action)
        self.app.updates = types.SimpleNamespace(load=lambda payload: None)

        with tempfile.TemporaryDirectory() as tmpdir:
            project_path = os.path.join(tmpdir, "example.osp")
            assets_path = os.path.join(tmpdir, "example_assets")
            thumbnail_root = os.path.join(assets_path, "thumbnail")
            os.makedirs(thumbnail_root, exist_ok=True)
            flat_thumb = os.path.join(thumbnail_root, "F1-8.png")
            with open(flat_thumb, "w", encoding="utf-8") as handle:
                handle.write("thumb")

            default_thumb_root = os.path.join(tmpdir, "default-thumbs")
            os.mkdir(default_thumb_root)

            with ExitStack() as stack:
                stack.enter_context(
                    patch.object(store, "new", lambda: setattr(store, "_data", project_data.copy()))
                )
                stack.enter_context(
                    patch.object(
                        store,
                        "read_from_file",
                        lambda file_path, path_mode="ignore": project_data.copy(),
                    )
                )
                stack.enter_context(patch.object(store, "check_if_paths_are_valid", lambda: None))
                stack.enter_context(patch.object(store, "add_to_recent_files", lambda file_path: None))
                stack.enter_context(patch.object(store, "upgrade_project_data_structures", lambda: None))
                stack.enter_context(patch.object(store, "get_profile", lambda **kwargs: object()))
                stack.enter_context(patch.object(store, "apply_default_audio_settings", lambda: None))
                stack.enter_context(patch("classes.project_data.get_assets_path", return_value=assets_path))
                stack.enter_context(patch("classes.project_data.info.get_default_path", return_value=default_thumb_root))
                ProjectDataStore.load(store, project_path, clear_thumbnails=True)

            self.assertFalse(os.path.exists(flat_thumb))
            self.assertTrue(os.path.exists(os.path.join(thumbnail_root, "F1", "8.png")))

    def test_upgrade_project_data_structures_migrates_25_crop_effect(self):
        store = make_store()
        self.app.project = types.SimpleNamespace(generate_id=lambda: "EFF-1")
        store._data = {
            "version": {"openshot-qt": "2.5.1", "libopenshot": "0.2.7"},
            "id": "T0",
            "clips": [{
                "id": "C1",
                "effects": [],
                "crop_x": {"Points": [{"co": {"Y": 0.25}}]},
                "crop_y": {"Points": [{"co": {"Y": 0.0}}]},
                "crop_width": {"Points": [{"co": {"Y": 0.75}}]},
                "crop_height": {"Points": [{"co": {"Y": 0.5}}]},
            }],
        }

        with patch.object(store, "generate_id", lambda digits=10: "NEW-PROJECT-ID"):
            ProjectDataStore.upgrade_project_data_structures(store)

        clip = store._data["clips"][0]
        self.assertNotIn("crop_x", clip)
        self.assertTrue(clip["effects"])
        effect = clip["effects"][0]
        self.assertEqual(effect["id"], "EFF-1")
        self.assertEqual(effect["x"]["Points"][0]["co"]["Y"], 0.25)
        self.assertEqual(effect["right"]["Points"][0]["co"]["Y"], 0.25)
        self.assertEqual(effect["bottom"]["Points"][0]["co"]["Y"], 0.5)
        self.assertEqual(store._data["id"], "NEW-PROJECT-ID")

    def test_add_to_recent_files_moves_existing_path_to_end(self):
        store = make_store()
        with tempfile.TemporaryDirectory() as tmpdir:
            one = os.path.join(tmpdir, "one.osp")
            two = os.path.join(tmpdir, "two.osp")
            three = os.path.join(tmpdir, "three.osp")
            self.app.settings.values["recent_projects"] = [one, two, three]

            ProjectDataStore.add_to_recent_files(store, two)

            self.assertEqual(
                self.app.settings.values["recent_projects"],
                [one, three, two],
            )
            self.assertTrue(self.app.settings.saved)

    def test_move_temp_paths_to_project_folder_copies_proxy_reader_into_project_assets(self):
        store = make_store()
        with tempfile.TemporaryDirectory() as tmpdir:
            proxy_root = os.path.join(tmpdir, "runtime-optimized")
            os.mkdir(proxy_root)
            proxy_file = os.path.join(proxy_root, "F1.mp4")
            with open(proxy_file, "wb") as handle:
                handle.write(b"proxy")

            project_path = os.path.join(tmpdir, "example.osp")
            asset_path = os.path.join(tmpdir, "example_assets")
            target_proxy_path = os.path.join(asset_path, "optimized")

            store._data = {
                "files": [
                    {
                        "id": "F1",
                        "path": os.path.join(tmpdir, "source.mp4"),
                        "proxy_reader": {
                            "id": "F1",
                            "path": proxy_file,
                        },
                    }
                ],
                "clips": [],
            }

            with ExitStack() as stack:
                stack.enter_context(patch("classes.project_data.get_assets_path", lambda path, create_paths=True: asset_path))
                stack.enter_context(patch("classes.project_data.info.PROXY_PATH", proxy_root))
                stack.enter_context(patch("classes.project_data.info.THUMBNAIL_PATH", os.path.join(tmpdir, "thumbs")))
                stack.enter_context(patch("classes.project_data.info.TITLE_PATH", os.path.join(tmpdir, "titles")))
                stack.enter_context(patch("classes.project_data.info.BLENDER_PATH", os.path.join(tmpdir, "blender")))
                stack.enter_context(patch("classes.project_data.info.PROTOBUF_DATA_PATH", os.path.join(tmpdir, "protobuf")))
                stack.enter_context(patch("classes.project_data.info.CLIPBOARD_PATH", os.path.join(tmpdir, "clipboard")))
                stack.enter_context(patch("classes.project_data.info.COMFYUI_OUTPUT_PATH", os.path.join(tmpdir, "comfy")))
                for folder in ("thumbs", "titles", "blender", "protobuf", "clipboard", "comfy"):
                    os.mkdir(os.path.join(tmpdir, folder))
                ProjectDataStore.move_temp_paths_to_project_folder(store, project_path)

            self.assertTrue(os.path.exists(os.path.join(target_proxy_path, "F1.mp4")))
            self.assertEqual(
                store._data["files"][0]["proxy_reader"]["path"],
                os.path.join(target_proxy_path, "F1.mp4"),
            )

    def test_load_migrates_legacy_proxy_folder_to_optimized(self):
        store = make_store()
        project_data = {
            "clips": [],
            "effects": [],
            "markers": [],
            "layers": [],
            "files": [{
                "id": "F1",
                "path": "/project/source.mp4",
                "proxy_reader": {"path": ""},
            }],
            "history": {"undo": [], "redo": []},
            "version": {"openshot-qt": "3.4.0", "libopenshot": "0.5.0"},
        }

        clear_waveform_action = DummyAction()
        self.app.window = types.SimpleNamespace(actionClearWaveformData=clear_waveform_action)
        self.app.updates = types.SimpleNamespace(load=lambda payload: None)

        with tempfile.TemporaryDirectory() as tmpdir:
            project_path = os.path.join(tmpdir, "example.osp")
            assets_path = os.path.join(tmpdir, "example_assets")
            legacy_proxy_root = os.path.join(assets_path, "proxies")
            os.makedirs(legacy_proxy_root, exist_ok=True)
            legacy_proxy = os.path.join(legacy_proxy_root, "F1.mp4")
            with open(legacy_proxy, "wb") as handle:
                handle.write(b"proxy")
            project_data["files"][0]["proxy_reader"]["path"] = legacy_proxy

            default_thumb_root = os.path.join(tmpdir, "default-thumbs")
            os.mkdir(default_thumb_root)

            with ExitStack() as stack:
                stack.enter_context(
                    patch.object(store, "new", lambda: setattr(store, "_data", project_data.copy()))
                )
                stack.enter_context(
                    patch.object(
                        store,
                        "read_from_file",
                        lambda file_path, path_mode="ignore": project_data.copy(),
                    )
                )
                stack.enter_context(patch.object(store, "check_if_paths_are_valid", lambda: None))
                stack.enter_context(patch.object(store, "add_to_recent_files", lambda file_path: None))
                stack.enter_context(patch.object(store, "upgrade_project_data_structures", lambda: None))
                stack.enter_context(patch.object(store, "get_profile", lambda **kwargs: object()))
                stack.enter_context(patch.object(store, "apply_default_audio_settings", lambda: None))
                stack.enter_context(patch("classes.project_data.get_assets_path", return_value=assets_path))
                stack.enter_context(patch("classes.project_data.info.get_default_path", return_value=default_thumb_root))
                ProjectDataStore.load(store, project_path, clear_thumbnails=True)

            expected_proxy = os.path.join(assets_path, "optimized", "F1.mp4")
            self.assertFalse(os.path.exists(legacy_proxy))
            self.assertTrue(os.path.exists(expected_proxy))
            self.assertEqual(store._data["files"][0]["proxy_reader"]["path"], expected_proxy)

    def test_upgrade_project_data_structures_migrates_tracker_alpha_and_parent(self):
        store = make_store()
        store._data = {
            "version": {"openshot-qt": "3.1.1", "libopenshot": "0.3.0"},
            "id": "P1",
            "clips": [
                {
                    "id": "PARENT",
                    "effects": [{
                        "name": "Tracker",
                        "display_box_text": {"Points": [{"co": {"Y": 0.25}}]},
                        "objects": {
                            "obj-1": {
                                "child_clip_id": "CHILD",
                                "background_alpha": {"Points": [{"co": {"Y": 0.2}}]},
                                "stroke_alpha": {"Points": [{"co": {"Y": 0.8}}]},
                            }
                        },
                    }],
                },
                {"id": "CHILD", "effects": []},
            ],
        }

        ProjectDataStore.upgrade_project_data_structures(store)

        effect = store._data["clips"][0]["effects"][0]
        tracked = effect["objects"]["obj-1"]
        self.assertEqual(effect["display_box_text"]["Points"][0]["co"]["Y"], 0.75)
        self.assertEqual(tracked["background_alpha"]["Points"][0]["co"]["Y"], 0.8)
        self.assertEqual(tracked["stroke_alpha"]["Points"][0]["co"]["Y"], 0.19999999999999996)
        self.assertEqual(store._data["clips"][1]["parentObjectId"], "obj-1")

    def test_check_if_paths_are_valid_updates_missing_file_and_syncs_clip_reader(self):
        store = make_store()
        old_path = "/missing/file.mp4"
        new_path = "/found/file.mp4"
        store._data = {
            "files": [{"id": "F1", "path": old_path}],
            "clips": [{"id": "C1", "file_id": "F1", "reader": {"path": old_path}}],
            "effects": [],
        }

        self.app.window = types.SimpleNamespace()

        with ExitStack() as stack:
            stack.enter_context(
                patch("classes.project_data.os.path.exists", side_effect=lambda p: p == new_path)
            )
            stack.enter_context(
                patch("classes.project_data.find_missing_file", return_value=(new_path, True, False))
            )
            ProjectDataStore.check_if_paths_are_valid(store)

        self.assertEqual(store._data["files"][0]["path"], new_path)
        self.assertEqual(store._data["clips"][0]["reader"]["path"], new_path)
        self.assertEqual(self.app.settings.default_paths[self.app.settings.actionType.IMPORT], new_path)

    def test_check_if_paths_are_valid_reuses_decision_for_duplicate_missing_effect_paths(self):
        store = make_store()
        missing_path = "/missing/shared-mask.svg"
        new_path = "/found/shared-mask.svg"
        store._data = {
            "files": [],
            "clips": [],
            "effects": [
                {"id": "E1", "resource": missing_path},
                {"id": "E2", "resource": missing_path},
            ],
        }

        self.app.window = types.SimpleNamespace()
        calls = []

        def fake_find_missing_file(path, prompt_state):
            calls.append(path)
            return new_path, True, False

        with ExitStack() as stack:
            stack.enter_context(
                patch("classes.project_data.os.path.exists", side_effect=lambda p: p == new_path)
            )
            stack.enter_context(
                patch("classes.project_data.find_missing_file", side_effect=fake_find_missing_file)
            )
            ProjectDataStore.check_if_paths_are_valid(store)

        self.assertEqual(calls, [missing_path])
        self.assertEqual(store._data["effects"][0]["resource"], new_path)
        self.assertEqual(store._data["effects"][1]["resource"], new_path)

    def test_check_if_paths_are_valid_removes_missing_effect_when_skipped(self):
        store = make_store()
        missing_path = "/missing/mask.svg"
        store._data = {
            "files": [],
            "clips": [],
            "effects": [{"id": "E1", "resource": missing_path}],
        }

        self.app.window = types.SimpleNamespace()

        with ExitStack() as stack:
            stack.enter_context(patch("classes.project_data.os.path.exists", return_value=False))
            stack.enter_context(
                patch("classes.project_data.find_missing_file", return_value=("", False, True))
            )
            ProjectDataStore.check_if_paths_are_valid(store)

        self.assertEqual(store._data["effects"], [])

    def test_check_if_paths_are_valid_reuses_skip_decision_for_duplicate_missing_paths(self):
        store = make_store()
        missing_path = "/missing/shared.svg"
        store._data = {
            "files": [],
            "clips": [],
            "effects": [
                {"id": "E1", "resource": missing_path},
                {"id": "E2", "resource": missing_path},
            ],
        }

        self.app.window = types.SimpleNamespace()
        calls = []

        def fake_find_missing_file(path, prompt_state):
            calls.append(path)
            prompt_state["last_skip"] = "all"
            return "", False, True

        with ExitStack() as stack:
            stack.enter_context(patch("classes.project_data.os.path.exists", return_value=False))
            stack.enter_context(
                patch("classes.project_data.find_missing_file", side_effect=fake_find_missing_file)
            )
            ProjectDataStore.check_if_paths_are_valid(store)

        self.assertEqual(calls, [missing_path])
        self.assertEqual(store._data["effects"], [])

    def test_check_if_paths_are_valid_repairs_missing_clip_effect_reader_path(self):
        store = make_store()
        missing_path = "/missing/effect-mask.svg"
        found_path = "/found/effect-mask.svg"
        store._data = {
            "files": [],
            "clips": [{
                "id": "C1",
                "effects": [{
                    "id": "E1",
                    "mask_reader": {"path": missing_path},
                }],
            }],
            "effects": [],
        }

        self.app.window = types.SimpleNamespace()

        with ExitStack() as stack:
            stack.enter_context(
                patch("classes.project_data.os.path.exists", side_effect=lambda p: p == found_path)
            )
            stack.enter_context(
                patch("classes.project_data.find_missing_file", return_value=(found_path, True, False))
            )
            ProjectDataStore.check_if_paths_are_valid(store)

        self.assertEqual(
            store._data["clips"][0]["effects"][0]["mask_reader"]["path"],
            found_path,
        )

    def test_move_temp_paths_to_project_folder_updates_title_file_and_clip_reader(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = os.path.abspath(tmpdir)
            old_title_dir = os.path.join(tmpdir, "working_title")
            old_thumb_dir = os.path.join(tmpdir, "working_thumb")
            old_blender_dir = os.path.join(tmpdir, "working_blender")
            old_proto_dir = os.path.join(tmpdir, "working_proto")
            old_clipboard_dir = os.path.join(tmpdir, "working_clipboard")
            old_comfy_dir = os.path.join(tmpdir, "working_comfy")
            for path in [
                old_title_dir, old_thumb_dir, old_blender_dir,
                old_proto_dir, old_clipboard_dir, old_comfy_dir,
            ]:
                os.mkdir(path)

            title_path = os.path.join(old_title_dir, "title.svg")
            with open(title_path, "w", encoding="utf-8") as handle:
                handle.write("<svg />")

            store = make_store()
            store._data = {
                "files": [{"id": "F1", "path": title_path}],
                "clips": [{"id": "C1", "file_id": "F1", "reader": {"path": title_path}, "effects": []}],
            }

            target_assets = os.path.join(tmpdir, "project_assets")
            project_file = os.path.join(tmpdir, "project.osp")

            with ExitStack() as stack:
                stack.enter_context(
                    patch("classes.project_data.get_assets_path", return_value=target_assets)
                )
                stack.enter_context(patch("classes.project_data.info.THUMBNAIL_PATH", old_thumb_dir))
                stack.enter_context(patch("classes.project_data.info.TITLE_PATH", old_title_dir))
                stack.enter_context(patch("classes.project_data.info.BLENDER_PATH", old_blender_dir))
                stack.enter_context(
                    patch("classes.project_data.info.PROTOBUF_DATA_PATH", old_proto_dir)
                )
                stack.enter_context(
                    patch("classes.project_data.info.CLIPBOARD_PATH", old_clipboard_dir)
                )
                stack.enter_context(
                    patch("classes.project_data.info.COMFYUI_OUTPUT_PATH", old_comfy_dir)
                )
                ProjectDataStore.move_temp_paths_to_project_folder(store, project_file)

            expected_title = os.path.join(target_assets, "title", "title.svg")
            self.assertEqual(store._data["files"][0]["path"], expected_title)
            self.assertEqual(store._data["clips"][0]["reader"]["path"], expected_title)
            self.assertTrue(os.path.exists(expected_title))

    def test_move_temp_paths_to_project_folder_copies_nested_thumbnail_layout(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = os.path.abspath(tmpdir)
            old_title_dir = os.path.join(tmpdir, "working_title")
            old_thumb_dir = os.path.join(tmpdir, "working_thumb")
            old_blender_dir = os.path.join(tmpdir, "working_blender")
            old_proto_dir = os.path.join(tmpdir, "working_proto")
            old_clipboard_dir = os.path.join(tmpdir, "working_clipboard")
            old_comfy_dir = os.path.join(tmpdir, "working_comfy")
            for path in [
                old_title_dir, old_thumb_dir, old_blender_dir,
                old_proto_dir, old_clipboard_dir, old_comfy_dir,
            ]:
                os.mkdir(path)

            nested_thumb_dir = os.path.join(old_thumb_dir, "F1")
            os.mkdir(nested_thumb_dir)
            nested_thumb_path = os.path.join(nested_thumb_dir, "8.png")
            with open(nested_thumb_path, "w", encoding="utf-8") as handle:
                handle.write("thumb")

            store = make_store()
            store._data = {"files": [], "clips": []}

            target_assets = os.path.join(tmpdir, "project_assets")
            project_file = os.path.join(tmpdir, "project.osp")

            with ExitStack() as stack:
                stack.enter_context(
                    patch("classes.project_data.get_assets_path", return_value=target_assets)
                )
                stack.enter_context(patch("classes.project_data.info.THUMBNAIL_PATH", old_thumb_dir))
                stack.enter_context(patch("classes.project_data.info.TITLE_PATH", old_title_dir))
                stack.enter_context(patch("classes.project_data.info.BLENDER_PATH", old_blender_dir))
                stack.enter_context(
                    patch("classes.project_data.info.PROTOBUF_DATA_PATH", old_proto_dir)
                )
                stack.enter_context(
                    patch("classes.project_data.info.CLIPBOARD_PATH", old_clipboard_dir)
                )
                stack.enter_context(
                    patch("classes.project_data.info.COMFYUI_OUTPUT_PATH", old_comfy_dir)
                )
                ProjectDataStore.move_temp_paths_to_project_folder(store, project_file)

            self.assertTrue(os.path.exists(os.path.join(target_assets, "thumbnail", "F1", "8.png")))

    def test_move_temp_paths_to_project_folder_moves_runtime_proxy_and_cleans_source(self):
        store = make_store()
        with tempfile.TemporaryDirectory() as tmpdir:
            proxy_root = os.path.join(tmpdir, "runtime-optimized")
            os.mkdir(proxy_root)
            proxy_file = os.path.join(proxy_root, "F1.mp4")
            with open(proxy_file, "wb") as handle:
                handle.write(b"proxy")

            project_path = os.path.join(tmpdir, "example.osp")
            asset_path = os.path.join(tmpdir, "example_assets")
            target_proxy_path = os.path.join(asset_path, "optimized")

            store._data = {
                "files": [
                    {
                        "id": "F1",
                        "path": os.path.join(tmpdir, "source.mp4"),
                        "proxy_reader": {
                            "id": "F1",
                            "path": proxy_file,
                        },
                    }
                ],
                "clips": [],
            }

            default_paths = {
                "PROXY_PATH": proxy_root,
                "THUMBNAIL_PATH": os.path.join(tmpdir, "thumbs"),
                "TITLE_PATH": os.path.join(tmpdir, "titles"),
                "BLENDER_PATH": os.path.join(tmpdir, "blender"),
                "PROTOBUF_DATA_PATH": os.path.join(tmpdir, "protobuf"),
                "CLIPBOARD_PATH": os.path.join(tmpdir, "clipboard"),
                "COMFYUI_OUTPUT_PATH": os.path.join(tmpdir, "comfy"),
            }

            with ExitStack() as stack:
                stack.enter_context(patch("classes.project_data.get_assets_path", lambda path, create_paths=True: asset_path))
                stack.enter_context(patch("classes.project_data.info.PROXY_PATH", proxy_root))
                stack.enter_context(patch("classes.project_data.info.THUMBNAIL_PATH", default_paths["THUMBNAIL_PATH"]))
                stack.enter_context(patch("classes.project_data.info.TITLE_PATH", default_paths["TITLE_PATH"]))
                stack.enter_context(patch("classes.project_data.info.BLENDER_PATH", default_paths["BLENDER_PATH"]))
                stack.enter_context(patch("classes.project_data.info.PROTOBUF_DATA_PATH", default_paths["PROTOBUF_DATA_PATH"]))
                stack.enter_context(patch("classes.project_data.info.CLIPBOARD_PATH", default_paths["CLIPBOARD_PATH"]))
                stack.enter_context(patch("classes.project_data.info.COMFYUI_OUTPUT_PATH", default_paths["COMFYUI_OUTPUT_PATH"]))
                stack.enter_context(
                    patch(
                        "classes.project_data.info.get_default_path",
                        side_effect=default_paths.get,
                    )
                )
                for name, folder in default_paths.items():
                    if name != "PROXY_PATH":
                        os.mkdir(folder)
                ProjectDataStore.move_temp_paths_to_project_folder(store, project_path)

            self.assertTrue(os.path.exists(os.path.join(target_proxy_path, "F1.mp4")))
            self.assertFalse(os.path.exists(proxy_file))
            self.assertEqual(
                store._data["files"][0]["proxy_reader"]["path"],
                os.path.join(target_proxy_path, "F1.mp4"),
            )

    def test_move_temp_paths_to_project_folder_removes_runtime_duplicates_when_target_exists(self):
        store = make_store()
        with tempfile.TemporaryDirectory() as tmpdir:
            title_root = os.path.join(tmpdir, "runtime-title")
            os.mkdir(title_root)
            source_title = os.path.join(title_root, "title.svg")
            with open(source_title, "w", encoding="utf-8") as handle:
                handle.write("<svg />")

            project_path = os.path.join(tmpdir, "example.osp")
            asset_path = os.path.join(tmpdir, "example_assets")
            target_title_path = os.path.join(asset_path, "title")
            os.makedirs(target_title_path, exist_ok=True)
            existing_title = os.path.join(target_title_path, "title.svg")
            with open(existing_title, "w", encoding="utf-8") as handle:
                handle.write("<svg />")

            store._data = {
                "files": [{"id": "F1", "path": source_title}],
                "clips": [{"id": "C1", "file_id": "F1", "reader": {"path": source_title}, "effects": []}],
            }

            default_paths = {
                "THUMBNAIL_PATH": os.path.join(tmpdir, "thumbs"),
                "TITLE_PATH": title_root,
                "BLENDER_PATH": os.path.join(tmpdir, "blender"),
                "PROTOBUF_DATA_PATH": os.path.join(tmpdir, "protobuf"),
                "CLIPBOARD_PATH": os.path.join(tmpdir, "clipboard"),
                "COMFYUI_OUTPUT_PATH": os.path.join(tmpdir, "comfy"),
                "PROXY_PATH": os.path.join(tmpdir, "optimized"),
            }

            with ExitStack() as stack:
                stack.enter_context(patch("classes.project_data.get_assets_path", lambda path, create_paths=True: asset_path))
                stack.enter_context(patch("classes.project_data.info.THUMBNAIL_PATH", default_paths["THUMBNAIL_PATH"]))
                stack.enter_context(patch("classes.project_data.info.TITLE_PATH", default_paths["TITLE_PATH"]))
                stack.enter_context(patch("classes.project_data.info.BLENDER_PATH", default_paths["BLENDER_PATH"]))
                stack.enter_context(patch("classes.project_data.info.PROTOBUF_DATA_PATH", default_paths["PROTOBUF_DATA_PATH"]))
                stack.enter_context(patch("classes.project_data.info.CLIPBOARD_PATH", default_paths["CLIPBOARD_PATH"]))
                stack.enter_context(patch("classes.project_data.info.COMFYUI_OUTPUT_PATH", default_paths["COMFYUI_OUTPUT_PATH"]))
                stack.enter_context(patch("classes.project_data.info.PROXY_PATH", default_paths["PROXY_PATH"]))
                stack.enter_context(
                    patch(
                        "classes.project_data.info.get_default_path",
                        side_effect=default_paths.get,
                    )
                )
                for name, folder in default_paths.items():
                    if name != "TITLE_PATH":
                        os.mkdir(folder)
                ProjectDataStore.move_temp_paths_to_project_folder(store, project_path)

            self.assertFalse(os.path.exists(source_title))
            self.assertTrue(os.path.exists(existing_title))
            self.assertEqual(store._data["files"][0]["path"], existing_title)
            self.assertEqual(store._data["clips"][0]["reader"]["path"], existing_title)

    def test_move_temp_paths_to_project_folder_updates_blender_and_protobuf_paths_together(self):
        store = make_store()
        with tempfile.TemporaryDirectory() as tmpdir:
            blender_root = os.path.join(tmpdir, "runtime-blender")
            protobuf_root = os.path.join(tmpdir, "runtime-protobuf")
            thumb_root = os.path.join(tmpdir, "thumbs")
            title_root = os.path.join(tmpdir, "titles")
            clipboard_root = os.path.join(tmpdir, "clipboard")
            comfy_root = os.path.join(tmpdir, "comfy")
            proxy_root = os.path.join(tmpdir, "optimized")
            for folder in [blender_root, protobuf_root, thumb_root, title_root, clipboard_root, comfy_root, proxy_root]:
                os.mkdir(folder)

            blender_job_root = os.path.join(blender_root, "0NHHRJD8L4")
            os.mkdir(blender_job_root)
            blender_asset = os.path.join(blender_job_root, "TitleFileName%04d.png")
            with open(blender_asset, "w", encoding="utf-8") as handle:
                handle.write("frame-seq")

            protobuf_asset = os.path.join(protobuf_root, "B5ONPQNB8X.data")
            with open(protobuf_asset, "w", encoding="utf-8") as handle:
                handle.write("tracker-data")

            project_path = os.path.join(tmpdir, "example.osp")
            asset_path = os.path.join(tmpdir, "example_assets")

            store._data = {
                "files": [
                    {"id": "F1", "path": blender_asset},
                ],
                "clips": [
                    {
                        "id": "C1",
                        "file_id": "F1",
                        "reader": {"path": blender_asset},
                        "effects": [
                            {"id": "E1", "protobuf_data_path": protobuf_asset},
                        ],
                    },
                ],
            }

            default_paths = {
                "THUMBNAIL_PATH": thumb_root,
                "TITLE_PATH": title_root,
                "BLENDER_PATH": blender_root,
                "PROTOBUF_DATA_PATH": protobuf_root,
                "CLIPBOARD_PATH": clipboard_root,
                "COMFYUI_OUTPUT_PATH": comfy_root,
                "PROXY_PATH": proxy_root,
            }

            with ExitStack() as stack:
                stack.enter_context(patch("classes.project_data.get_assets_path", lambda path, create_paths=True: asset_path))
                stack.enter_context(patch("classes.project_data.info.THUMBNAIL_PATH", default_paths["THUMBNAIL_PATH"]))
                stack.enter_context(patch("classes.project_data.info.TITLE_PATH", default_paths["TITLE_PATH"]))
                stack.enter_context(patch("classes.project_data.info.BLENDER_PATH", default_paths["BLENDER_PATH"]))
                stack.enter_context(patch("classes.project_data.info.PROTOBUF_DATA_PATH", default_paths["PROTOBUF_DATA_PATH"]))
                stack.enter_context(patch("classes.project_data.info.CLIPBOARD_PATH", default_paths["CLIPBOARD_PATH"]))
                stack.enter_context(patch("classes.project_data.info.COMFYUI_OUTPUT_PATH", default_paths["COMFYUI_OUTPUT_PATH"]))
                stack.enter_context(patch("classes.project_data.info.PROXY_PATH", default_paths["PROXY_PATH"]))
                stack.enter_context(
                    patch(
                        "classes.project_data.info.get_default_path",
                        side_effect=default_paths.get,
                    )
                )
                ProjectDataStore.move_temp_paths_to_project_folder(store, project_path)

            expected_blender = os.path.join(asset_path, "blender", "0NHHRJD8L4", "TitleFileName%04d.png")
            expected_protobuf = os.path.join(asset_path, "protobuf_data", "B5ONPQNB8X.data")
            self.assertEqual(store._data["files"][0]["path"], expected_blender)
            self.assertEqual(store._data["clips"][0]["reader"]["path"], expected_blender)
            self.assertEqual(
                store._data["clips"][0]["effects"][0]["protobuf_data_path"],
                expected_protobuf,
            )
            self.assertTrue(os.path.exists(expected_blender))
            self.assertTrue(os.path.exists(expected_protobuf))



"""
 @file
 @brief Painter for the custom scrollbars.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QRectF
from PyQt5.QtGui import QBrush, QColor, QPainter

from .base import BasePainter


class ScrollbarPainter(BasePainter):
    """Draw horizontal and vertical scrollbars."""

    def update_theme(self):
        handle = getattr(self.w.theme, "scrollbar_handle", QColor())
        track = getattr(self.w.theme, "scrollbar_track", QColor())
        if not handle.isValid():
            handle = QColor("#4b92ad")
        if not track.isValid():
            track = QColor("#000")
        self.handle_brush = QBrush(handle)
        self.track_brush = QBrush(track)

    def paint(self, painter: QPainter):
        # Horizontal scrollbar
        sb = self.w.scroll_bar_rect
        if not sb.isNull():
            track = QRectF(
                self.w.track_name_width,
                self.w.height() - self.w.scroll_bar_thickness,
                self.w.width() - self.w.track_name_width - self.w.scroll_bar_thickness,
                self.w.scroll_bar_thickness,
            )
            painter.fillRect(track, self.track_brush)
            painter.fillRect(sb, self.handle_brush)

        # Vertical scrollbar
        sbv = getattr(self.w, "v_scroll_bar_rect", QRectF())
        if not sbv.isNull():
            track = QRectF(
                self.w.width() - self.w.scroll_bar_thickness,
                self.w.ruler_height,
                self.w.scroll_bar_thickness,
                self.w.height() - self.w.ruler_height - self.w.scroll_bar_thickness,
            )
            painter.fillRect(track, self.track_brush)
            painter.fillRect(sbv, self.handle_brush)



"""
 @file
 @brief This file can easily query Clips, Files, and other project data
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """
import copy
import json
import os

import openshot

from classes import info
from classes.app import get_app
from classes.path_utils import absolute_media_path, media_paths_equal


class QueryObject:
    """ This class allows one or more project data objects to be queried """

    # Cache detached project objects per update version
    _cache_version = None
    _cache = {}

    def __init__(self):
        """ Constructor """

        self.id = None  # Unique ID of object
        self.key = None  # Key path to object in project data
        self.data = None  # Data dictionary of object
        self.parent = None  # Only used with effects (who belong to clips)
        self.type = "insert"  # Type of operation needed to save

    def save(self, OBJECT_TYPE):
        """ Save the object back to the project data store """

        # Insert or Update this data into the project data store
        if not self.id and self.type == "insert":

            # Insert record, and Generate id
            self.id = get_app().project.generate_id()

            # save id in data (if attribute found)
            self.data["id"] = json.loads(json.dumps(self.id))

            # Set key (if needed)
            if not self.key:
                self.key = json.loads(json.dumps(OBJECT_TYPE.object_key))
                self.key.append({"id": self.id})

            # Insert into project data
            get_app().updates.insert(json.loads(json.dumps(OBJECT_TYPE.object_key)), json.loads(json.dumps(self.data)))

            # Mark record as 'update' now... so another call to this method won't insert it again
            self.type = "update"

        elif self.id and self.type == "update":

            # Update existing project data
            get_app().updates.update(self.key, self.data)

    def delete(self, OBJECT_TYPE):
        """ Delete the object from the project data store """

        # Delete if object found and not pending insert
        if self.id and self.type == "update":
            # Delete from project data store
            get_app().updates.delete(self.key)
            self.type = "delete"

    def title(self):
        """ Get the translated display title of this item """
        # Needs to be overwritten in each derived class
        return None

    @classmethod
    def _get_cached_child(cls, OBJECT_TYPE, child):
        """Return a cached, detached copy of child, clearing cache when project changes"""
        updates = get_app().updates
        current_version = getattr(updates, "data_version", 0)

        if cls._cache_version != current_version:
            cls._cache = {}
            cls._cache_version = current_version

        object_cache = cls._cache.setdefault(OBJECT_TYPE.object_name, {})
        child_id = child.get("id")

        # Cache deep copies by id; reuse within the same project version
        cached = object_cache.get(child_id)
        if cached is None:
            cached = copy.deepcopy(child)
            object_cache[child_id] = cached
        return cached

    def filter(OBJECT_TYPE, **kwargs):
        """ Take any arguments given as filters, and find a list of matching objects """

        # Get a list of all objects of this type
        parent = get_app().project.get(OBJECT_TYPE.object_key)

        if not parent:
            return []

        matching_objects = []

        # Loop through all children objects
        for child in parent:

            # Protect against non-iterable/subscriptables
            if not child:
                continue

            # Loop through all kwargs (and look for matches)
            match = True
            for key, value in kwargs.items():

                if key in child and child[key] != value:
                    match = False
                    break

                # Intersection Position
                if key == "intersect" and (
                    child.get("position", 0) > value
                    or child.get("position", 0) + (child.get("end", 0) - child.get("start", 0)) < value
                ):
                    match = False


            # Add matched record
            if match:
                object = OBJECT_TYPE()
                object.id = child["id"]
                object.key = [OBJECT_TYPE.object_name, {"id": object.id}]
                object.data = QueryObject._get_cached_child(OBJECT_TYPE, child)
                object.type = "update"
                matching_objects.append(object)

        # Return matching objects
        return matching_objects

    def get(OBJECT_TYPE, **kwargs):
        """ Take any arguments given as filters, and find the first matching object """

        # Look for matching objects
        matching_objects = QueryObject.filter(OBJECT_TYPE, **kwargs)

        if matching_objects:
            return matching_objects[0]
        else:
            return None


class Clip(QueryObject):
    """ This class allows Clips to be queried, updated, and deleted from the project data. """
    object_name = "clips"  # Derived classes should define this
    object_key = [object_name]  # Derived classes should define this also

    def save(self):
        """ Save the object back to the project data store """
        super().save(Clip)

    def delete(self):
        """ Delete the object from the project data store """
        super().delete(Clip)

    def filter(**kwargs):
        """ Take any arguments given as filters, and find a list of matching objects """
        return QueryObject.filter(Clip, **kwargs)

    def get(**kwargs):
        """ Take any arguments given as filters, and find the first matching object """
        return QueryObject.get(Clip, **kwargs)

    def title(self):
        """ Get the translated display title of this item """
        path = self.data.get("reader", {}).get("path")
        file_id = self.data.get("file_id") or self.data.get("reader", {}).get("id")
        if file_id:
            file = File.get(id=file_id)
            if file:
                name = str(file.data.get("name", "")).strip()
                if name:
                    return name
        return os.path.basename(path)

class Transition(QueryObject):
    """ This class allows Transitions (i.e. timeline effects) to be queried, updated, and deleted from the project data. """
    object_name = "effects"  # Derived classes should define this
    object_key = [object_name]  # Derived classes should define this also

    def save(self):
        """ Save the object back to the project data store """
        super().save(Transition)

    def delete(self):
        """ Delete the object from the project data store """
        super().delete(Transition)

    def filter(**kwargs):
        """ Take any arguments given as filters, and find a list of matching objects """
        return QueryObject.filter(Transition, **kwargs)

    def get(**kwargs):
        """ Take any arguments given as filters, and find the first matching object """
        return QueryObject.get(Transition, **kwargs)

    def title(self):
        """ Get the translated display title of this item """
        path = self.data.get("reader", {}).get("path")
        if not path:
            return None
        fileBaseName = os.path.splitext(os.path.basename(path))[0]

        # split the name into parts (looking for a number)
        suffix_number = None
        name_parts = fileBaseName.split("_")
        if name_parts[-1].isdigit():
            suffix_number = name_parts[-1]
        # get name of transition
        item_name = fileBaseName.replace("_", " ").capitalize()

        # replace suffix number with placeholder (if any)
        if suffix_number:
            item_name = item_name.replace(suffix_number, "%s")
            item_name = get_app()._tr(item_name) % suffix_number
        else:
            item_name = get_app()._tr(item_name)
        return item_name


class File(QueryObject):
    """ This class allows Files to be queried, updated, and deleted from the project data. """
    object_name = "files"  # Derived classes should define this
    object_key = [object_name]  # Derived classes should define this also

    def save(self):
        """ Save the object back to the project data store """
        super().save(File)

    def delete(self):
        """ Delete the object from the project data store """
        super().delete(File)

    def filter(**kwargs):
        """ Take any arguments given as filters, and find a list of matching objects """
        return QueryObject.filter(File, **kwargs)

    def get(**kwargs):
        """ Take any arguments given as filters, and find the first matching object """
        file_path = kwargs.pop("path", None)
        if file_path is None:
            return QueryObject.get(File, **kwargs)

        matching_objects = QueryObject.filter(File, **kwargs)
        for obj in matching_objects:
            if media_paths_equal(obj.data.get("path"), file_path):
                return obj
        return None

    def absolute_path(self):
        """ Get absolute file path of file """

        file_path = self.data.get("path")
        resolved = absolute_media_path(file_path)
        if resolved:
            return resolved
        return file_path or ""

    def relative_path(self):
        """ Get relative path (based on the current working directory) """

        file_path = self.absolute_path()
        # Convert path to relative (based on current working directory of Python)
        return os.path.relpath(file_path, info.CWD)

    def profile(self):
        """ Get the profile of the file """
        # Load file Json into Profile object
        file_profile = openshot.Profile()
        file_profile.SetJson(json.dumps(self.data))

        if file_profile.info.display_ratio.num == 1 and file_profile.info.display_ratio.den == 1:
            # Some audio / image files have inaccurate DAR - calculate from size and pixel ratio
            file_profile.info.display_ratio = openshot.Fraction(round(file_profile.info.width * file_profile.info.pixel_ratio.ToFloat()), file_profile.info.height)
            file_profile.info.display_ratio.Reduce()

        # Iterate through all possible profiles
        for profile_folder in [info.USER_PROFILES_PATH, info.PROFILES_PATH]:
            for file in reversed(sorted(os.listdir(profile_folder))):
                profile_path = os.path.join(profile_folder, file)
                if os.path.isdir(profile_path):
                    continue
                try:
                    # Load Profile
                    profile = openshot.Profile(profile_path)
                    if profile == file_profile:
                        return profile
                except RuntimeError as e:
                    pass
        return file_profile


class Marker(QueryObject):
    """ This class allows Markers to be queried, updated, and deleted from the project data. """
    object_name = "markers"  # Derived classes should define this
    object_key = [object_name]  # Derived classes should define this also

    def save(self):
        """ Save the object back to the project data store """
        super().save(Marker)

    def delete(self):
        """ Delete the object from the project data store """
        super().delete(Marker)

    def filter(**kwargs):
        """ Take any arguments given as filters, and find a list of matching objects """
        return QueryObject.filter(Marker, **kwargs)

    def get(**kwargs):
        """ Take any arguments given as filters, and find the first matching object """
        return QueryObject.get(Marker, **kwargs)


class Track(QueryObject):
    """ This class allows Tracks to be queried, updated, and deleted from the project data. """
    object_name = "layers"  # Derived classes should define this
    object_key = [object_name]  # Derived classes should define this also

    def save(self):
        """ Save the object back to the project data store """
        super().save(Track)

    def delete(self):
        """ Delete the object from the project data store """
        super().delete(Track)

    def filter(**kwargs):
        """ Take any arguments given as filters, and find a list of matching objects """
        return QueryObject.filter(Track, **kwargs)

    def get(**kwargs):
        """ Take any arguments given as filters, and find the first matching object """
        return QueryObject.get(Track, **kwargs)

    def __lt__(self, other):
        return self.data.get('number', 0) < other.data.get('number', 0)

    def __gt__(self, other):
        return self.data.get('number', 0) > other.data.get('number', 0)


class Effect(QueryObject):
    """ This class allows Effects to be queried, updated, and deleted from the project data. """
    object_name = "effects"  # Derived classes should define this
    object_key = [object_name]  # Derived classes should define this also

    def save(self):
        """ Save the object back to the project data store """
        super().save(Effect)

    def delete(self):
        """ Delete the object from the project data store """
        super().delete(Effect)

    def filter(**kwargs):
        """ Take any arguments given as filters, and find a list of matching objects """

        # Get a list of clips
        clips = get_app().project.get("clips")
        matching_objects = []

        # Loop through all clips
        if clips:
            for clip in clips:
                # Loop through all effects
                if "effects" in clip:
                    for child in clip["effects"]:

                        # Loop through all kwargs (and look for matches)
                        match = True
                        for key, value in kwargs.items():
                            if key in child and child[key] != value:
                                match = False
                                break

                        # Add matched record
                        if match:
                            object = Effect()
                            object.id = child["id"]
                            object.key = ["clips", {"id": clip["id"]}, "effects", {"id": object.id}]
                            object.data = json.loads(json.dumps(child))
                            object.type = "update"
                            object.parent = clip
                            matching_objects.append(object)

        # Return matching objects
        return matching_objects

    def title(self):
        """ Get the translated display title of this item """
        return self.data.get("name") or self.data.get("type")

    def get(**kwargs):
        """ Take any arguments given as filters, and find the first matching object """
        # Look for matching objects
        matching_objects = Effect.filter(**kwargs)

        if matching_objects:
            return matching_objects[0]
        else:
            return None
